# ClusterShell.CLI.Nodeset test suite
# Written by S. Thiell

"""Unit test for CLI.Nodeset"""

import os
import random
from textwrap import dedent
import unittest

from TLib import *
from ClusterShell.CLI.Nodeset import main

from ClusterShell.NodeUtils import GroupResolverConfig
from ClusterShell.NodeSet import set_std_group_resolver, \
                                 set_std_group_resolver_config


class CLINodesetTestBase(unittest.TestCase):
    """Base unit test class for testing CLI/Nodeset.py"""

    def _nodeset_t(self, args, stdin, expected_stdout, expected_rc=0,
                   expected_stderr=None):
        CLI_main(self, main, ['nodeset'] + args, stdin, expected_stdout,
                 expected_rc, expected_stderr)


class CLINodesetTest(CLINodesetTestBase):
    """Unit test class for testing CLI/Nodeset.py"""

    def _battery_count(self, args):
        self._nodeset_t(args + ["--count", ""], None, b"0\n")
        self._nodeset_t(args + ["--count", "foo"], None, b"1\n")
        self._nodeset_t(args + ["--count", "foo", "bar"], None, b"2\n")
        self._nodeset_t(args + ["--count", "foo", "foo"], None, b"1\n")
        self._nodeset_t(args + ["--count", "foo", "foo", "bar"], None, b"2\n")
        self._nodeset_t(args + ["--count", "foo[0]"], None, b"1\n")
        self._nodeset_t(args + ["--count", "foo[2]"], None, b"1\n")
        self._nodeset_t(args + ["--count", "foo[1,2]"], None, b"2\n")
        self._nodeset_t(args + ["--count", "foo[1-2]"], None, b"2\n")
        self._nodeset_t(args + ["--count", "foo[1,2]", "foo[1-2]"], None, b"2\n")
        self._nodeset_t(args + ["--count", "foo[1-200,245-394]"], None, b"350\n")
        self._nodeset_t(args + ["--count", "foo[395-442]", "foo[1-200,245-394]"], None, b"398\n")
        self._nodeset_t(args + ["--count", "foo[395-442]", "foo", "foo[1-200,245-394]"], None, b"399\n")
        self._nodeset_t(args + ["--count", "foo[395-442]", "foo", "foo[0-200,245-394]"], None, b"400\n")
        self._nodeset_t(args + ["--count", "foo[395-442]", "bar3,bar24", "foo[1-200,245-394]"], None, b"400\n")
        # from stdin: use string not bytes as input because CLI/Nodeset.py works in text mode
        self._nodeset_t(args + ["--count"], "\n", b"0\n")
        self._nodeset_t(args + ["--count"], "foo\n", b"1\n")
        self._nodeset_t(args + ["--count"], "foo\nbar\n", b"2\n")
        self._nodeset_t(args + ["--count"], "foo\nfoo\n", b"1\n")
        self._nodeset_t(args + ["--count"], "foo\nfoo\nbar\n", b"2\n")
        self._nodeset_t(args + ["--count"], "foo[0]\n", b"1\n")
        self._nodeset_t(args + ["--count"], "foo[2]\n", b"1\n")
        self._nodeset_t(args + ["--count"], "foo[1,2]\n", b"2\n")
        self._nodeset_t(args + ["--count"], "foo[1-2]\n", b"2\n")
        self._nodeset_t(args + ["--count"], "foo[1,2]\nfoo[1-2]\n", b"2\n")
        self._nodeset_t(args + ["--count"], "foo[1-200,245-394]\n", b"350\n")
        self._nodeset_t(args + ["--count"], "foo[395-442]\nfoo[1-200,245-394]\n", b"398\n")
        self._nodeset_t(args + ["--count"], "foo[395-442]\nfoo\nfoo[1-200,245-394]\n", b"399\n")
        self._nodeset_t(args + ["--count"], "foo[395-442]\nfoo\nfoo[0-200,245-394]\n", b"400\n")
        self._nodeset_t(args + ["--count"], "foo[395-442]\nbar3,bar24\nfoo[1-200,245-394]\n", b"400\n")

    def test_001_count(self):
        """test nodeset --count"""
        self._battery_count([])
        self._battery_count(["--autostep=1"])
        self._battery_count(["--autostep=2"])
        self._battery_count(["--autostep=5"])
        self._battery_count(["--autostep=auto"])
        self._battery_count(["--autostep=0%"])
        self._battery_count(["--autostep=50%"])
        self._battery_count(["--autostep=100%"])

    def test_002_count_intersection(self):
        """test nodeset --count --intersection"""
        self._nodeset_t(["--count", "foo", "--intersection", "bar"], None, b"0\n")
        self._nodeset_t(["--count", "foo", "--intersection", "foo"], None, b"1\n")
        self._nodeset_t(["--count", "foo", "--intersection", "foo", "-i", "bar"], None, b"0\n")
        self._nodeset_t(["--count", "foo[0]", "--intersection", "foo0"], None, b"1\n")
        self._nodeset_t(["--count", "foo[2]", "--intersection", "foo"], None, b"0\n")
        self._nodeset_t(["--count", "foo[1,2]", "--intersection", "foo[1-2]"], None, b"2\n")
        self._nodeset_t(["--count", "foo[395-442]", "--intersection", "foo[1-200,245-394]"], None, b"0\n")
        self._nodeset_t(["--count", "foo[395-442]", "--intersection", "foo", "-i", "foo[1-200,245-394]"], None, b"0\n")
        self._nodeset_t(["--count", "foo[395-442]", "-i", "foo", "-i", "foo[0-200,245-394]"], None, b"0\n")
        self._nodeset_t(["--count", "foo[395-442]", "--intersection", "bar3,bar24", "-i", "foo[1-200,245-394]"], None, b"0\n")
        # multiline args (#394)
        self._nodeset_t(["--count", "foo[1,2]", "-i", "foo1\nfoo2"], None, b"2\n")
        self._nodeset_t(["--count", "foo[1,2]", "-i", "foo1\nfoo2", "foo3\nfoo4"], None, b"4\n")

    def test_003_count_intersection_stdin(self):
        """test nodeset --count --intersection (stdin)"""
        self._nodeset_t(["--count", "--intersection", "bar"], "foo\n", b"0\n")
        self._nodeset_t(["--count", "--intersection", "foo"], "foo\n", b"1\n")
        self._nodeset_t(["--count", "--intersection", "foo", "-i", "bar"], "foo\n", b"0\n")
        self._nodeset_t(["--count", "--intersection", "foo0"], "foo[0]\n", b"1\n")
        self._nodeset_t(["--count", "--intersection", "foo"], "foo[2]\n", b"0\n")
        self._nodeset_t(["--count", "--intersection", "foo[1-2]"], "foo[1,2]\n", b"2\n")
        self._nodeset_t(["--count", "--intersection", "foo[1-200,245-394]"], "foo[395-442]\n", b"0\n")
        self._nodeset_t(["--count", "--intersection", "foo", "-i", "foo[1-200,245-394]"], "foo[395-442]\n", b"0\n")
        self._nodeset_t(["--count", "-i", "foo", "-i", "foo[0-200,245-394]"], "foo[395-442]\n", b"0\n")
        self._nodeset_t(["--count", "--intersection", "bar3,bar24", "-i", "foo[1-200,245-394]"], "foo[395-442]\n", b"0\n")

    def _battery_fold(self, args):
        self._nodeset_t(args + ["--fold", ""], None, b"\n")
        self._nodeset_t(args + ["--fold", "foo"], None, b"foo\n")
        self._nodeset_t(args + ["--fold", "foo", "bar"], None, b"bar,foo\n")
        self._nodeset_t(args + ["--fold", "foo", "foo"], None, b"foo\n")
        self._nodeset_t(args + ["--fold", "foo", "foo", "bar"], None, b"bar,foo\n")
        self._nodeset_t(args + ["--fold", "foo[0]"], None, b"foo0\n")
        self._nodeset_t(args + ["--fold", "foo[2]"], None, b"foo2\n")
        self._nodeset_t(args + ["--fold", "foo[1,2]"], None, b"foo[1-2]\n")
        self._nodeset_t(args + ["--fold", "foo[1-2]"], None, b"foo[1-2]\n")
        self._nodeset_t(args + ["--fold", "foo[1,2]", "foo[1-2]"], None, b"foo[1-2]\n")
        self._nodeset_t(args + ["--fold", "foo[1-200,245-394]"], None, b"foo[1-200,245-394]\n")
        self._nodeset_t(args + ["--fold", "foo[395-442]", "foo[1-200,245-394]"], None, b"foo[1-200,245-442]\n")
        self._nodeset_t(args + ["--fold", "foo[395-442]", "foo", "foo[1-200,245-394]"], None, b"foo,foo[1-200,245-442]\n")
        self._nodeset_t(args + ["--fold", "foo[395-442]", "foo", "foo[0-200,245-394]"], None, b"foo,foo[0-200,245-442]\n")
        self._nodeset_t(args + ["--fold", "foo[395-442]", "bar3,bar24", "foo[1-200,245-394]"], None, b"bar[3,24],foo[1-200,245-442]\n")
        # multiline arg (#394)
        self._nodeset_t(args + ["--fold", "foo3\nfoo1\nfoo2\nbar"], None, b"bar,foo[1-3]\n")
        self._nodeset_t(args + ["--fold", "foo3\n\n\nfoo1\n\nfoo2\n\n"], None, b"foo[1-3]\n")
        # stdin
        self._nodeset_t(args + ["--fold"], "\n", b"\n")
        self._nodeset_t(args + ["--fold"], "foo\n", b"foo\n")
        self._nodeset_t(args + ["--fold"], "foo\nbar\n", b"bar,foo\n")
        self._nodeset_t(args + ["--fold"], "foo\nfoo\n", b"foo\n")
        self._nodeset_t(args + ["--fold"], "foo\nfoo\nbar\n", b"bar,foo\n")
        self._nodeset_t(args + ["--fold"], "foo[0]\n", b"foo0\n")
        self._nodeset_t(args + ["--fold"], "foo[2]\n", b"foo2\n")
        self._nodeset_t(args + ["--fold"], "foo[1,2]\n", b"foo[1-2]\n")
        self._nodeset_t(args + ["--fold"], "foo[1-2]\n", b"foo[1-2]\n")
        self._nodeset_t(args + ["--fold"], "foo[1,2]\nfoo[1-2]\n", b"foo[1-2]\n")
        self._nodeset_t(args + ["--fold"], "foo[1-200,245-394]\n", b"foo[1-200,245-394]\n")
        self._nodeset_t(args + ["--fold"], "foo[395-442]\nfoo[1-200,245-394]\n", b"foo[1-200,245-442]\n")
        self._nodeset_t(args + ["--fold"], "foo[395-442]\nfoo\nfoo[1-200,245-394]\n", b"foo,foo[1-200,245-442]\n")
        self._nodeset_t(args + ["--fold"], "foo[395-442]\nfoo\nfoo[0-200,245-394]\n", b"foo,foo[0-200,245-442]\n")
        self._nodeset_t(args + ["--fold"], "foo[395-442]\nbar3,bar24\nfoo[1-200,245-394]\n", b"bar[3,24],foo[1-200,245-442]\n")

    def test_004_fold(self):
        """test nodeset --fold"""
        self._battery_fold([])
        self._battery_fold(["--autostep=3"])
        # --autostep=auto (1.7)
        self._battery_fold(["--autostep=auto"])
        self._battery_count(["--autostep=0%"])
        self._battery_count(["--autostep=50%"])
        self._battery_count(["--autostep=100%"])

    def test_005_fold_autostep(self):
        """test nodeset --fold --autostep=X"""
        self._nodeset_t(["--autostep=2", "-f", "foo0", "foo2", "foo4", "foo6"], None, b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=2", "-f", "foo4", "foo2", "foo0", "foo6"], None, b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=3", "-f", "foo0", "foo2", "foo4", "foo6"], None, b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=4", "-f", "foo0", "foo2", "foo4", "foo6"], None, b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=5", "-f", "foo0", "foo2", "foo4", "foo6"], None, b"foo[0,2,4,6]\n")
        self._nodeset_t(["--autostep=auto", "-f", "foo0", "foo2", "foo4", "foo6"], None, b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=auto", "-f", "foo4", "foo2", "foo0", "foo6"], None, b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=auto", "-f", "foo4", "foo2", "foo0", "foo2", "foo6"], None, b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=auto", "-f", "foo4", "foo2", "foo0", "foo5", "foo6"], None, b"foo[0,2,4-6]\n")
        self._nodeset_t(["--autostep=auto", "-f", "foo4", "foo2", "foo0", "foo9", "foo6"], None, b"foo[0,2,4,6,9]\n")
        self._nodeset_t(["--autostep=75%", "-f", "foo0", "foo2", "foo4", "foo6"], None, b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=75%", "-f", "foo4", "foo2", "foo0", "foo6"], None, b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=80%", "-f", "foo4", "foo2", "foo0", "foo2", "foo6"], None, b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=80%", "-f", "foo4", "foo2", "foo0", "foo5", "foo6"], None, b"foo[0,2,4-6]\n")
        self._nodeset_t(["--autostep=80%", "-f", "foo4", "foo2", "foo0", "foo9", "foo6"], None, b"foo[0-6/2,9]\n")
        self._nodeset_t(["--autostep=81%", "-f", "foo4", "foo2", "foo0", "foo9", "foo6"], None, b"foo[0,2,4,6,9]\n")
        self._nodeset_t(["--autostep=100%", "-f", "foo4", "foo2", "foo0", "foo9", "foo6"], None, b"foo[0,2,4,6,9]\n")

    def test_006_expand(self):
        """test nodeset --expand"""
        self._nodeset_t(["--expand", ""], None, b"\n")
        self._nodeset_t(["--expand", "foo"], None, b"foo\n")
        self._nodeset_t(["--expand", "foo", "bar"], None, b"bar foo\n")
        self._nodeset_t(["--expand", "foo", "foo"], None, b"foo\n")
        self._nodeset_t(["--expand", "foo[0]"], None, b"foo0\n")
        self._nodeset_t(["--expand", "foo[2]"], None, b"foo2\n")
        self._nodeset_t(["--expand", "foo[1,2]"], None, b"foo1 foo2\n")
        self._nodeset_t(["--expand", "foo[1-2]"], None, b"foo1 foo2\n")
        self._nodeset_t(["--expand", "foo[1-2],bar"], None, b"bar foo1 foo2\n")

    def test_007_expand_stdin(self):
        """test nodeset --expand (stdin)"""
        self._nodeset_t(["--expand"], "\n", b"\n")
        self._nodeset_t(["--expand"], "foo\n", b"foo\n")
        self._nodeset_t(["--expand"], "foo\nbar\n", b"bar foo\n")
        self._nodeset_t(["--expand"], "foo\nfoo\n", b"foo\n")
        self._nodeset_t(["--expand"], "foo[0]\n", b"foo0\n")
        self._nodeset_t(["--expand"], "foo[2]\n", b"foo2\n")
        self._nodeset_t(["--expand"], "foo[1,2]\n", b"foo1 foo2\n")
        self._nodeset_t(["--expand"], "foo[1-2]\n", b"foo1 foo2\n")
        self._nodeset_t(["--expand"], "foo[1-2],bar\n", b"bar foo1 foo2\n")

    def test_008_expand_separator(self):
        """test nodeset --expand -S"""
        self._nodeset_t(["--expand", "-S", ":", "foo"], None, b"foo\n")
        self._nodeset_t(["--expand", "-S", ":", "foo", "bar"], None, b"bar:foo\n")
        self._nodeset_t(["--expand", "--separator", ":", "foo", "bar"], None, b"bar:foo\n")
        self._nodeset_t(["--expand", "--separator=:", "foo", "bar"], None, b"bar:foo\n")
        self._nodeset_t(["--expand", "-S", ":", "foo", "foo"], None, b"foo\n")
        self._nodeset_t(["--expand", "-S", ":", "foo[0]"], None, b"foo0\n")
        self._nodeset_t(["--expand", "-S", ":", "foo[2]"], None, b"foo2\n")
        self._nodeset_t(["--expand", "-S", ":", "foo[1,2]"], None, b"foo1:foo2\n")
        self._nodeset_t(["--expand", "-S", ":", "foo[1-2]"], None, b"foo1:foo2\n")
        self._nodeset_t(["--expand", "-S", " ", "foo[1-2]"], None, b"foo1 foo2\n")
        self._nodeset_t(["--expand", "-S", ",", "foo[1-2],bar"], None, b"bar,foo1,foo2\n")
        self._nodeset_t(["--expand", "-S", "uuu", "foo[1-2],bar"], None, b"baruuufoo1uuufoo2\n")
        self._nodeset_t(["--expand", "-S", "\\n", "foo[1-2]"], None, b"foo1\nfoo2\n")
        self._nodeset_t(["--expand", "-S", "\n", "foo[1-2]"], None, b"foo1\nfoo2\n")

    def test_009_fold_xor(self):
        """test nodeset --fold --xor"""
        self._nodeset_t(["--fold", "foo", "-X", "bar"], None, b"bar,foo\n")
        self._nodeset_t(["--fold", "foo", "-X", "foo"], None, b"\n")
        self._nodeset_t(["--fold", "foo[1,2]", "-X", "foo[1-2]"], None, b"\n")
        self._nodeset_t(["--fold", "foo[1-10]", "-X", "foo[5-15]"], None, b"foo[1-4,11-15]\n")
        self._nodeset_t(["--fold", "foo[395-442]", "-X", "foo[1-200,245-394]"], None, b"foo[1-200,245-442]\n")
        self._nodeset_t(["--fold", "foo[395-442]", "-X", "foo", "-X", "foo[1-200,245-394]"], None, b"foo,foo[1-200,245-442]\n")
        self._nodeset_t(["--fold", "foo[395-442]", "-X", "foo", "-X", "foo[0-200,245-394]"], None, b"foo,foo[0-200,245-442]\n")
        self._nodeset_t(["--fold", "foo[395-442]", "-X", "bar3,bar24", "-X", "foo[1-200,245-394]"], None, b"bar[3,24],foo[1-200,245-442]\n")
        # multiline args (#394)
        self._nodeset_t(["--fold", "foo[1-10]", "-X", "foo5\nfoo6\nfoo7"], None, b"foo[1-4,8-10]\n")
        self._nodeset_t(["--fold", "foo[1-10]", "-X", "foo5\nfoo6\nfoo7", "foo5\nfoo6"], None, b"foo[1-6,8-10]\n")

    def test_010_fold_xor_stdin(self):
        """test nodeset --fold --xor (stdin)"""
        self._nodeset_t(["--fold", "-X", "bar"], "foo\n", b"bar,foo\n")
        self._nodeset_t(["--fold", "-X", "foo"], "foo\n", b"\n")
        self._nodeset_t(["--fold", "-X", "foo[1-2]"], "foo[1,2]\n", b"\n")
        self._nodeset_t(["--fold", "-X", "foo[5-15]"], "foo[1-10]\n", b"foo[1-4,11-15]\n")
        self._nodeset_t(["--fold", "-X", "foo[1-200,245-394]"], "foo[395-442]\n", b"foo[1-200,245-442]\n")
        self._nodeset_t(["--fold", "-X", "foo", "-X", "foo[1-200,245-394]"], "foo[395-442]\n", b"foo,foo[1-200,245-442]\n")
        self._nodeset_t(["--fold", "-X", "foo", "-X", "foo[0-200,245-394]"], "foo[395-442]\n", b"foo,foo[0-200,245-442]\n")
        self._nodeset_t(["--fold", "-X", "bar3,bar24", "-X", "foo[1-200,245-394]"], "foo[395-442]\n", b"bar[3,24],foo[1-200,245-442]\n")
        # using stdin for -X
        self._nodeset_t(["-f", "foo[2-4]", "-X", "-"], "foo4 foo5 foo6\n", b"foo[2-3,5-6]\n")
        self._nodeset_t(["-f", "-X", "-", "foo[1-6]"], "foo4 foo5 foo6\n",
                        b"foo[1-6]\n", 0,
                        b"WARNING: empty left operand for set operation\n")

    def test_011_fold_exclude(self):
        """test nodeset --fold --exclude"""
        # Empty result
        self._nodeset_t(["--fold", "foo", "-x", "foo"], None, b"\n")
        # With no range
        self._nodeset_t(["--fold", "foo,bar", "-x", "foo"], None, b"bar\n")
        # Normal with range
        self._nodeset_t(["--fold", "foo[0-5]", "-x", "foo[0-10]"], None, b"\n")
        self._nodeset_t(["--fold", "foo[0-10]", "-x", "foo[0-5]"], None, b"foo[6-10]\n")
        # Do no change
        self._nodeset_t(["--fold", "foo[6-10]", "-x", "bar[0-5]"], None, b"foo[6-10]\n")
        self._nodeset_t(["--fold", "foo[0-10]", "foo[13-18]", "--exclude", "foo[5-10,15]"], None, b"foo[0-4,13-14,16-18]\n")
        # multiline args (#394)
        self._nodeset_t(["--fold", "foo[0-5]", "-x", "foo0\nfoo9\nfoo3\nfoo2\nfoo1"], None, b"foo[4-5]\n")
        self._nodeset_t(["--fold", "foo[0-5]", "-x", "foo0\nfoo9\nfoo3\nfoo2\nfoo1", "foo5\nfoo6"], None, b"foo[4-6]\n")

    def test_012_fold_exclude_stdin(self):
        """test nodeset --fold --exclude (stdin)"""
        # Empty result
        self._nodeset_t(["--fold", "-x", "foo"], "", b"\n", 0,
                        b"WARNING: empty left operand for set operation\n")
        self._nodeset_t(["--fold", "-x", "foo"], "\n", b"\n", 0,
                        b"WARNING: empty left operand for set operation\n")
        self._nodeset_t(["--fold", "-x", "foo"], "foo\n", b"\n")
        # With no range
        self._nodeset_t(["--fold", "-x", "foo"], "foo,bar\n", b"bar\n")
        # Normal with range
        self._nodeset_t(["--fold", "-x", "foo[0-10]"], "foo[0-5]\n", b"\n")
        self._nodeset_t(["--fold", "-x", "foo[0-5]"], "foo[0-10]\n", b"foo[6-10]\n")
        # Do no change
        self._nodeset_t(["--fold", "-x", "bar[0-5]"], "foo[6-10]\n", b"foo[6-10]\n")
        self._nodeset_t(["--fold", "--exclude", "foo[5-10,15]"], "foo[0-10]\nfoo[13-18]\n", b"foo[0-4,13-14,16-18]\n")
        # using stdin for -x
        self._nodeset_t(["-f", "foo[1-6]", "-x", "-"], "foo4 foo5 foo6\n", b"foo[1-3]\n")
        self._nodeset_t(["-f", "-x", "-", "foo[1-6]"], "foo4 foo5 foo6\n",
                        b"foo[1-6]\n", 0,
                        b"WARNING: empty left operand for set operation\n")

    def test_013_fold_intersection(self):
        """test nodeset --fold --intersection"""
        # Empty result
        self._nodeset_t(["--fold", "foo", "-i", "foo"], None, b"foo\n")
        # With no range
        self._nodeset_t(["--fold", "foo,bar", "--intersection", "foo"], None, b"foo\n")
        # Normal with range
        self._nodeset_t(["--fold", "foo[0-5]", "-i", "foo[0-10]"], None, b"foo[0-5]\n")
        self._nodeset_t(["--fold", "foo[0-10]", "-i", "foo[0-5]"], None, b"foo[0-5]\n")
        self._nodeset_t(["--fold", "foo[6-10]", "-i", "bar[0-5]"], None, b"\n")
        self._nodeset_t(["--fold", "foo[0-10]", "foo[13-18]", "-i", "foo[5-10,15]"], None, b"foo[5-10,15]\n")
        # numerical bracket folding (#228)
        self._nodeset_t(["--fold", "node123[1-2]", "-i", "node1232"], None, b"node1232\n")
        self._nodeset_t(["--fold", "node023[1-2]0", "-i", "node02320"], None, b"node02320\n")
        self._nodeset_t(["--fold", "node023[1-2]0-ipmi2", "-i", "node02320-ipmi2"], None, b"node02320-ipmi2\n")
        self._nodeset_t(["--fold", "-i", "foo", "foo"], None, b"foo\n", 0,
                        b"WARNING: empty left operand for set operation\n")

    def test_014_fold_intersection_stdin(self):
        """test nodeset --fold --intersection (stdin)"""
        # Empty result
        self._nodeset_t(["--fold", "--intersection", "foo"], "", b"\n", 0,
                        b"WARNING: empty left operand for set operation\n")
        self._nodeset_t(["--fold", "--intersection", "foo"], "\n", b"\n", 0,
                        b"WARNING: empty left operand for set operation\n")
        self._nodeset_t(["--fold", "-i", "foo"], "foo\n", b"foo\n")
        # With no range
        self._nodeset_t(["--fold", "-i", "foo"], "foo,bar\n", b"foo\n")
        # Normal with range
        self._nodeset_t(["--fold", "-i", "foo[0-10]"], "foo[0-5]\n", b"foo[0-5]\n")
        self._nodeset_t(["--fold", "-i", "foo[0-5]"], "foo[0-10]\n", b"foo[0-5]\n")
        # Do no change
        self._nodeset_t(["--fold", "-i", "bar[0-5]"], "foo[6-10]\n", b"\n")
        self._nodeset_t(["--fold", "-i", "foo[5-10,15]"], "foo[0-10]\nfoo[13-18]\n", b"foo[5-10,15]\n")
        # using stdin for -i
        self._nodeset_t(["-f", "foo[1-6]", "-i", "-"], "foo4 foo5 foo6\n", b"foo[4-6]\n")
        self._nodeset_t(["-f", "-i", "-", "foo[1-6]"], "foo4 foo5 foo6\n",
                        b"foo[1-6]\n", 0,
                        b"WARNING: empty left operand for set operation\n")
        # numerical bracket folding (#228)
        self._nodeset_t(["--fold", "-i", "node123[1-2]"], "node1232\n", b"node1232\n")
        self._nodeset_t(["--fold", "-i", "node023[1-2]0"], "node02320\n", b"node02320\n")
        self._nodeset_t(["--fold", "-i", "node023[1-2]0-ipmi2"], "node02320-ipmi2\n", b"node02320-ipmi2\n")

    def test_015_rangeset(self):
        """test nodeset --rangeset"""
        self._nodeset_t(["--fold", "--rangeset", "1,2"], None, b"1-2\n")
        self._nodeset_t(["--expand", "-R", "1-2"], None, b"1 2\n")
        self._nodeset_t(["--fold", "-R", "1-2", "-X", "2-3"], None, b"1,3\n")

    def test_016_rangeset_stdin(self):
        """test nodeset --rangeset (stdin)"""
        self._nodeset_t(["--fold", "--rangeset"], "1,2\n", b"1-2\n")
        self._nodeset_t(["--expand", "-R"], "1-2\n", b"1 2\n")
        self._nodeset_t(["--fold", "-R", "-X", "2-3"], "1-2\n", b"1,3\n")

    def test_017_stdin(self):
        """test nodeset - (stdin)"""
        self._nodeset_t(["-f", "-"], "foo\n", b"foo\n")
        self._nodeset_t(["-f", "-"], "foo1 foo2 foo3\n", b"foo[1-3]\n")
        self._nodeset_t(["--autostep=2", "-f"], "foo0 foo2 foo4 foo6\n", b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=auto", "-f"], "foo0 foo2 foo4 foo6\n", b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=100%", "-f"], "foo0 foo2 foo4 foo6\n", b"foo[0-6/2]\n")
        self._nodeset_t(["--autostep=0%", "-f"], "foo0 foo2 foo4 foo6\n", b"foo[0-6/2]\n")

    def test_018_split(self):
        """test nodeset --split"""
        self._nodeset_t(["--split=2", "-f", "bar"], None, b"bar\n")
        self._nodeset_t(["--split", "2", "-f", "foo,bar"], None, b"bar\nfoo\n")
        self._nodeset_t(["--split", "2", "-e", "foo", "bar", "bur", "oof", "gcc"], None, b"bar bur foo\ngcc oof\n")
        self._nodeset_t(["--split=2", "-f", "foo[2-9]"], None, b"foo[2-5]\nfoo[6-9]\n")
        self._nodeset_t(["--split=2", "-f", "foo[2-3,7]", "bar9"], None, b"bar9,foo2\nfoo[3,7]\n")
        self._nodeset_t(["--split=3", "-f", "foo[2-9]"], None, b"foo[2-4]\nfoo[5-7]\nfoo[8-9]\n")
        self._nodeset_t(["--split=1", "-f", "foo2", "foo3"], None, b"foo[2-3]\n")
        self._nodeset_t(["--split=4", "-f", "foo[2-3]"], None, b"foo2\nfoo3\n")
        self._nodeset_t(["--split=4", "-f", "foo3", "foo2"], None, b"foo2\nfoo3\n")
        self._nodeset_t(["--split=2", "-e", "foo[2-9]"], None, b"foo2 foo3 foo4 foo5\nfoo6 foo7 foo8 foo9\n")
        self._nodeset_t(["--split=3", "-e", "foo[2-9]"], None, b"foo2 foo3 foo4\nfoo5 foo6 foo7\nfoo8 foo9\n")
        self._nodeset_t(["--split=1", "-e", "foo3", "foo2"], None, b"foo2 foo3\n")
        self._nodeset_t(["--split=4", "-e", "foo[2-3]"], None, b"foo2\nfoo3\n")
        self._nodeset_t(["--split=4", "-e", "foo2", "foo3"], None, b"foo2\nfoo3\n")
        self._nodeset_t(["--split=2", "-c", "foo2", "foo3"], None, b"1\n1\n")

    def test_019_contiguous(self):
        """test nodeset --contiguous"""
        self._nodeset_t(["--contiguous", "-f", "bar"], None, b"bar\n")
        self._nodeset_t(["--contiguous", "-f", "foo,bar"], None, b"bar\nfoo\n")
        self._nodeset_t(["--contiguous", "-f", "foo", "bar", "bur", "oof", "gcc"], None, b"bar\nbur\nfoo\ngcc\noof\n")
        self._nodeset_t(["--contiguous", "-e", "foo", "bar", "bur", "oof", "gcc"], None, b"bar\nbur\nfoo\ngcc\noof\n")
        self._nodeset_t(["--contiguous", "-f", "foo2"], None, b"foo2\n")
        self._nodeset_t(["--contiguous", "-R", "-f", "2"], None, b"2\n")
        self._nodeset_t(["--contiguous", "-f", "foo[2-9]"], None, b"foo[2-9]\n")
        self._nodeset_t(["--contiguous", "-f", "foo[2-3,7]", "bar9"], None, b"bar9\nfoo[2-3]\nfoo7\n")
        self._nodeset_t(["--contiguous", "-R", "-f", "2-3,7", "9"], None, b"2-3\n7\n9\n")
        self._nodeset_t(["--contiguous", "-f", "foo2", "foo3"], None, b"foo[2-3]\n")
        self._nodeset_t(["--contiguous", "-f", "foo3", "foo2"], None, b"foo[2-3]\n")
        self._nodeset_t(["--contiguous", "-f", "foo3", "foo1"], None, b"foo1\nfoo3\n")
        self._nodeset_t(["--contiguous", "-f", "foo[1-5/2]", "foo7"], None, b"foo1\nfoo3\nfoo5\nfoo7\n")

    def test_020_slice(self):
        """test nodeset -I/--slice"""
        self._nodeset_t(["--slice=0", "-f", "bar"], None, b"bar\n")
        self._nodeset_t(["--slice=0", "-e", "bar"], None, b"bar\n")
        self._nodeset_t(["--slice=1", "-f", "bar"], None, b"\n")
        self._nodeset_t(["--slice=0-1", "-f", "bar"], None, b"bar\n")
        self._nodeset_t(["-I0", "-f", "bar[34-68,89-90]"], None, b"bar34\n")
        self._nodeset_t(["-R", "-I0", "-f", "34-68,89-90"], None, b"34\n")
        self._nodeset_t(["-I 0", "-f", "bar[34-68,89-90]"], None, b"bar34\n")
        self._nodeset_t(["-I 0", "-e", "bar[34-68,89-90]"], None, b"bar34\n")
        self._nodeset_t(["-I 0-3", "-f", "bar[34-68,89-90]"], None, b"bar[34-37]\n")
        self._nodeset_t(["-I 0-3", "-f", "bar[34-68,89-90]", "-x", "bar34"], None, b"bar[35-38]\n")
        self._nodeset_t(["-I 0-3", "-f", "bar[34-68,89-90]", "-x", "bar35"], None, b"bar[34,36-38]\n")
        self._nodeset_t(["-I 0-3", "-e", "bar[34-68,89-90]"], None, b"bar34 bar35 bar36 bar37\n")
        self._nodeset_t(["-I 3,1,0,2", "-f", "bar[34-68,89-90]"], None, b"bar[34-37]\n")
        self._nodeset_t(["-I 1,3,7,10,16,20,30,34-35,37", "-f", "bar[34-68,89-90]"], None, b"bar[35,37,41,44,50,54,64,68,89]\n")
        self._nodeset_t(["-I 8", "-f", "bar[34-68,89-90]"], None, b"bar42\n")
        self._nodeset_t(["-I 8-100", "-f", "bar[34-68,89-90]"], None, b"bar[42-68,89-90]\n")
        self._nodeset_t(["-I 0-100", "-f", "bar[34-68,89-90]"], None, b"bar[34-68,89-90]\n")
        self._nodeset_t(["-I 8-100/2", "-f", "bar[34-68,89-90]"], None, b"bar[42,44,46,48,50,52,54,56,58,60,62,64,66,68,90]\n")
        self._nodeset_t(["--autostep=2", "-I 8-100/2", "-f", "bar[34-68,89-90]"], None, b"bar[42-68/2,90]\n")
        self._nodeset_t(["--autostep=93%", "-I 8-100/2", "-f", "bar[34-68,89-90]"], None, b"bar[42-68/2,90]\n")
        self._nodeset_t(["--autostep=94%", "-I 8-100/2", "-f", "bar[34-68,89-90]"], None, b"bar[42,44,46,48,50,52,54,56,58,60,62,64,66,68,90]\n")
        self._nodeset_t(["--autostep=auto", "-I 8-100/2", "-f", "bar[34-68,89-90]"], None, b"bar[42,44,46,48,50,52,54,56,58,60,62,64,66,68,90]\n")
        self._nodeset_t(["--autostep=auto", "-I 8-100/2", "-f", "bar[34-68]"], None, b"bar[42-68/2]\n")
        self._nodeset_t(["--autostep=100%", "-I 8-100/2", "-f", "bar[34-68]"], None, b"bar[42-68/2]\n")

    def test_021_slice_stdin(self):
        """test nodeset -I/--slice (stdin)"""
        self._nodeset_t(["--slice=0", "-f"], "bar\n", b"bar\n")
        self._nodeset_t(["--slice=0", "-e"], "bar\n", b"bar\n")
        self._nodeset_t(["--slice=1", "-f"], "bar\n", b"\n")
        self._nodeset_t(["--slice=0-1", "-f"], "bar\n", b"bar\n")
        self._nodeset_t(["-I0", "-f"], "bar[34-68,89-90]\n", b"bar34\n")
        self._nodeset_t(["-R", "-I0", "-f"], "34-68,89-90\n", b"34\n")
        self._nodeset_t(["-I 0", "-f"], "bar[34-68,89-90]\n", b"bar34\n")
        self._nodeset_t(["-I 0", "-e"], "bar[34-68,89-90]\n", b"bar34\n")
        self._nodeset_t(["-I 0-3", "-f"], "bar[34-68,89-90]\n", b"bar[34-37]\n")
        self._nodeset_t(["-I 0-3", "-f", "-x", "bar34"], "bar[34-68,89-90]\n", b"bar[35-38]\n")
        self._nodeset_t(["-I 0-3", "-f", "-x", "bar35"], "bar[34-68,89-90]\n", b"bar[34,36-38]\n")
        self._nodeset_t(["-I 0-3", "-e"], "bar[34-68,89-90]\n", b"bar34 bar35 bar36 bar37\n")
        self._nodeset_t(["-I 3,1,0,2", "-f"], "bar[34-68,89-90]\n", b"bar[34-37]\n")
        self._nodeset_t(["-I 1,3,7,10,16,20,30,34-35,37", "-f"], "bar[34-68,89-90]\n", b"bar[35,37,41,44,50,54,64,68,89]\n")
        self._nodeset_t(["-I 8", "-f"], "bar[34-68,89-90]\n", b"bar42\n")
        self._nodeset_t(["-I 8-100", "-f"], "bar[34-68,89-90]\n", b"bar[42-68,89-90]\n")
        self._nodeset_t(["-I 0-100", "-f"], "bar[34-68,89-90]\n", b"bar[34-68,89-90]\n")
        self._nodeset_t(["-I 8-100/2", "-f"], "bar[34-68,89-90]\n", b"bar[42,44,46,48,50,52,54,56,58,60,62,64,66,68,90]\n")
        self._nodeset_t(["--autostep=2", "-I 8-100/2", "-f"], "bar[34-68,89-90]\n", b"bar[42-68/2,90]\n")
        self._nodeset_t(["--autostep=93%", "-I 8-100/2", "-f"], "bar[34-68,89-90]\n", b"bar[42-68/2,90]\n")
        self._nodeset_t(["--autostep=93.33%", "-I 8-100/2", "-f"], "bar[34-68,89-90]\n", b"bar[42-68/2,90]\n")
        self._nodeset_t(["--autostep=94%", "-I 8-100/2", "-f"], "bar[34-68,89-90]\n", b"bar[42,44,46,48,50,52,54,56,58,60,62,64,66,68,90]\n")
        self._nodeset_t(["--autostep=auto", "-I 8-100/2", "-f"], "bar[34-68,89-90]\n", b"bar[42,44,46,48,50,52,54,56,58,60,62,64,66,68,90]\n")
        self._nodeset_t(["--autostep=2", "-I 8-100/2", "-f"], "bar[34-68]\n", b"bar[42-68/2]\n")

    def test_022_output_format(self):
        """test nodeset -O"""
        self._nodeset_t(["--expand", "--output-format", "/path/%s/", "foo"], None, b"/path/foo/\n")
        self._nodeset_t(["--expand", "-O", "/path/%s/", "-S", ":", "foo"], None, b"/path/foo/\n")
        self._nodeset_t(["--expand", "-O", "/path/%s/", "foo[2]"], None, b"/path/foo2/\n")
        self._nodeset_t(["--expand", "-O", "%s-ib0", "foo[1-4]"], None, b"foo1-ib0 foo2-ib0 foo3-ib0 foo4-ib0\n")
        self._nodeset_t(["--expand", "-O", "%s-ib0", "-S", ":", "foo[1-4]"], None, b"foo1-ib0:foo2-ib0:foo3-ib0:foo4-ib0\n")
        self._nodeset_t(["--fold", "-O", "%s-ipmi", "foo", "bar"], None, b"bar-ipmi,foo-ipmi\n")
        self._nodeset_t(["--fold", "-O", "%s-ib0", "foo1", "foo2"], None, b"foo[1-2]-ib0\n")
        self._nodeset_t(["--fold", "-O", "%s-ib0", "foo1", "foo2", "bar1", "bar2"], None, b"bar[1-2]-ib0,foo[1-2]-ib0\n")
        self._nodeset_t(["--fold", "-O", "%s-ib0", "--autostep=auto", "foo[1-9/2]"], None, b"foo[1-9/2]-ib0\n")
        self._nodeset_t(["--fold", "-O", "%s-ib0", "--autostep=6", "foo[1-9/2]"], None, b"foo[1,3,5,7,9]-ib0\n")
        self._nodeset_t(["--fold", "-O", "%s-ib0", "--autostep=5", "foo[1-9/2]"], None, b"foo[1-9/2]-ib0\n")
        self._nodeset_t(["--count", "-O", "result-%s", "foo1", "foo2"], None, b"result-2\n")
        self._nodeset_t(["--contiguous", "-O", "%s-ipmi", "-f", "foo[2-3,7]", "bar9"], None, b"bar9-ipmi\nfoo[2-3]-ipmi\nfoo7-ipmi\n")
        self._nodeset_t(["--split=2", "-O", "%s-ib", "-e", "foo[2-9]"], None, b"foo2-ib foo3-ib foo4-ib foo5-ib\nfoo6-ib foo7-ib foo8-ib foo9-ib\n")
        self._nodeset_t(["--split=3", "-O", "hwm-%s", "-f", "foo[2-9]"], None, b"hwm-foo[2-4]\nhwm-foo[5-7]\nhwm-foo[8-9]\n")
        self._nodeset_t(["-I0", "-O", "{%s}", "-f", "bar[34-68,89-90]"], None, b"{bar34}\n")
        # RangeSet mode (-R)
        self._nodeset_t(["--fold", "-O", "{%s}", "--rangeset", "1,2"], None, b"{1-2}\n")
        self._nodeset_t(["--expand", "-O", "{%s}", "-R", "1-2"], None, b"{1} {2}\n")
        self._nodeset_t(["--fold", "-O", "{%s}", "-R", "1-2", "-X", "2-3"], None, b"{1,3}\n")
        self._nodeset_t(["--fold", "-O", "{%s}", "-S", ":", "--rangeset", "1,2"], None, b"{1-2}\n")
        self._nodeset_t(["--expand", "-O", "{%s}", "-S", ":", "-R", "1-2"], None, b"{1}:{2}\n")
        self._nodeset_t(["--fold", "-O", "{%s}", "-S", ":", "-R", "1-2", "-X", "2-3"], None, b"{1,3}\n")
        self._nodeset_t(["-R", "-I0", "-O", "{%s}", "-f", "34-68,89-90"], None, b"{34}\n")

    def test_023_axis(self):
        """test nodeset folding with --axis"""
        self._nodeset_t(["--axis=0", "-f", "bar"], None, b"bar\n")
        self._nodeset_t(["--axis=1", "-f", "bar"], None, b"bar\n")
        self._nodeset_t(["--axis=1", "-R", "-f", "1,2,3"], None, None, 2,
                        b"--axis option is only supported when folding nodeset\n")
        self._nodeset_t(["--axis=1", "-e", "bar"], None, None, 2,
                        b"--axis option is only supported when folding nodeset\n")

        # 1D and 2D nodeset: fold along axis 0 only
        self._nodeset_t(["--axis=1", "-f", "comp-[1-2]-[1-3],login-[1-2]"], None,
                        b'comp-[1-2]-1,comp-[1-2]-2,comp-[1-2]-3,login-[1-2]\n')
        # 1D and 2D nodeset: fold along axis 1 only
        self._nodeset_t(["--axis=2", "-f", "comp-[1-2]-[1-3],login-[1-2]"], None,
                        b'comp-1-[1-3],comp-2-[1-3],login-1,login-2\n')
        # 1D and 2D nodeset: fold along last axis only
        self._nodeset_t(["--axis=-1", "-f", "comp-[1-2]-[1-3],login-[1-2]"], None,
                        b'comp-1-[1-3],comp-2-[1-3],login-[1-2]\n')

        # test for a common case
        ndnodes = []
        for ib in range(2):
            for idx in range(500):
                ndnodes.append("node%d-ib%d" % (idx, ib))
        random.shuffle(ndnodes)

        self._nodeset_t(["--axis=1", "-f"] + ndnodes, None,
                        b"node[0-499]-ib0,node[0-499]-ib1\n")

        exp_result = []
        for idx in range(500):
            exp_result.append("node%d-ib[0-1]" % idx)

        exp_result_str = '%s\n' % ','.join(exp_result)

        self._nodeset_t(["--axis=2", "-f"] + ndnodes, None,
                        exp_result_str.encode())

        # 4D test
        ndnodes = ["c-1-2-3-4", "c-2-2-3-4", "c-3-2-3-4", "c-5-5-5-5",
                   "c-5-7-5-5", "c-5-9-5-5", "c-5-11-5-5", "c-9-8-8-08",
                   "c-9-8-8-09"]
        self._nodeset_t(["--axis=1", "-f"] + ndnodes, None,
                        b"c-5-5-5-5,c-5-7-5-5,c-5-9-5-5,c-5-11-5-5,c-[1-3]-2-3-4,c-9-8-8-08,c-9-8-8-09\n")
        self._nodeset_t(["--axis=2", "-f"] + ndnodes, None,
                        b"c-5-[5,7,9,11]-5-5,c-1-2-3-4,c-2-2-3-4,c-3-2-3-4,c-9-8-8-08,c-9-8-8-09\n")
        self._nodeset_t(["--axis=3", "-f"] + ndnodes, None,
                        b"c-5-5-5-5,c-5-7-5-5,c-5-9-5-5,c-5-11-5-5,c-1-2-3-4,c-2-2-3-4,c-3-2-3-4,c-9-8-8-08,c-9-8-8-09\n")
        self._nodeset_t(["--axis=4", "-f"] + ndnodes, None,
                        b"c-5-5-5-5,c-5-7-5-5,c-5-9-5-5,c-5-11-5-5,c-1-2-3-4,c-2-2-3-4,c-3-2-3-4,c-9-8-8-[08-09]\n")

        self._nodeset_t(["--axis=1-2", "-f"] + ndnodes, None,
                        b"c-5-[5,7,9,11]-5-5,c-[1-3]-2-3-4,c-9-8-8-08,c-9-8-8-09\n")
        self._nodeset_t(["--axis=2-3", "-f"] + ndnodes, None,
                        b"c-5-[5,7,9,11]-5-5,c-1-2-3-4,c-2-2-3-4,c-3-2-3-4,c-9-8-8-08,c-9-8-8-09\n")
        self._nodeset_t(["--axis=3-4", "-f"] + ndnodes, None,
                        b"c-5-5-5-5,c-5-7-5-5,c-5-9-5-5,c-5-11-5-5,c-1-2-3-4,c-2-2-3-4,c-3-2-3-4,c-9-8-8-[08-09]\n")
        self._nodeset_t(["--axis=1-3", "-f"] + ndnodes, None,
                        b"c-5-[5,7,9,11]-5-5,c-[1-3]-2-3-4,c-9-8-8-08,c-9-8-8-09\n")
        self._nodeset_t(["--axis=2-4", "-f"] + ndnodes, None,
                        b"c-5-[5,7,9,11]-5-5,c-1-2-3-4,c-2-2-3-4,c-3-2-3-4,c-9-8-8-[08-09]\n")

        self._nodeset_t(["--axis=1-4", "-f"] + ndnodes, None,
                        b"c-5-[5,7,9,11]-5-5,c-[1-3]-2-3-4,c-9-8-8-[08-09]\n")
        self._nodeset_t(["-f"] + ndnodes, None,
                        b"c-5-[5,7,9,11]-5-5,c-[1-3]-2-3-4,c-9-8-8-[08-09]\n")

        # a case where axis and autostep are working
        self._nodeset_t(["--autostep=4", "--axis=1-2", "-f"] + ndnodes, None,
                        b"c-5-[5-11/2]-5-5,c-[1-3]-2-3-4,c-9-8-8-08,c-9-8-8-09\n")

    def test_024_axis_stdin(self):
        """test nodeset folding with --axis (stdin)"""
        self._nodeset_t(["--axis=0", "-f"], "bar\n", b"bar\n")
        self._nodeset_t(["--axis=1", "-f"], "bar\n", b"bar\n")
        self._nodeset_t(["--axis=1", "-R", "-f"], "1,2,3", None, 2,
                        b"--axis option is only supported when folding nodeset\n")
        self._nodeset_t(["--axis=1", "-e"], "bar\n", None, 2,
                        b"--axis option is only supported when folding nodeset\n")

        # 1D and 2D nodeset: fold along axis 0 only
        self._nodeset_t(["--axis=1", "-f"], "comp-[1-2]-[1-3],login-[1-2]\n",
                        b'comp-[1-2]-1,comp-[1-2]-2,comp-[1-2]-3,login-[1-2]\n')
        # 1D and 2D nodeset: fold along axis 1 only
        self._nodeset_t(["--axis=2", "-f"], "comp-[1-2]-[1-3],login-[1-2]\n",
                        b'comp-1-[1-3],comp-2-[1-3],login-1,login-2\n')
        # 1D and 2D nodeset: fold along last axis only
        self._nodeset_t(["--axis=-1", "-f"], "comp-[1-2]-[1-3],login-[1-2]\n",
                        b'comp-1-[1-3],comp-2-[1-3],login-[1-2]\n')

        # test for a common case
        ndnodes = []
        for ib in range(2):
            for idx in range(500):
                ndnodes.append("node%d-ib%d" % (idx, ib))
        random.shuffle(ndnodes)

        self._nodeset_t(["--axis=1", "-f"], '\n'.join(ndnodes) + '\n',
                        b"node[0-499]-ib0,node[0-499]-ib1\n")

        exp_result = []
        for idx in range(500):
            exp_result.append("node%d-ib[0-1]" % idx)

        exp_result_str = '%s\n' % ','.join(exp_result)

        self._nodeset_t(["--axis=2", "-f"], '\n'.join(ndnodes) + '\n',
                        exp_result_str.encode())

    def test_025_pick(self):
        """test nodeset --pick"""
        for num in range(1, 100):
            self._nodeset_t(["--count", "--pick", str(num), "foo[1-100]"],
                            None, str(num).encode() + b'\n')
            self._nodeset_t(["--count", "--pick", str(num), "-R", "1-100"],
                            None, str(num).encode() + b'\n')


class CLINodesetGroupResolverTest1(CLINodesetTestBase):
    """Unit test class for testing CLI/Nodeset.py with custom Group Resolver"""

    def setUp(self):
        # Special tests that require a default group source set
        #
        # The temporary file needs to be persistent during the tests
        # because GroupResolverConfig does lazy init, this is why we
        # use an instance variable self.f
        #
        self.f = make_temp_file(dedent("""
            [Main]
            default: local

            [local]
            map: echo example[1-100]
            all: echo example[1-1000]
            list: echo foo bar moo
            """).encode())
        set_std_group_resolver(GroupResolverConfig(self.f.name))

    def tearDown(self):
        set_std_group_resolver(None)
        self.f = None  # used to release temp file

    def test_022_list(self):
        """test nodeset --list"""
        self._nodeset_t(["--list"], None, b"@bar\n@foo\n@moo\n")
        self._nodeset_t(["-ll"], None, b"@bar example[1-100]\n@foo example[1-100]\n@moo example[1-100]\n")
        self._nodeset_t(["-lll"], None, b"@bar example[1-100] 100\n@foo example[1-100] 100\n@moo example[1-100] 100\n")
        self._nodeset_t(["-l", "example[4,95]", "example5"], None, b"@bar\n@foo\n@moo\n")
        self._nodeset_t(["-ll", "example[4,95]", "example5"], None, b"@bar example[4-5,95]\n@foo example[4-5,95]\n@moo example[4-5,95]\n")
        self._nodeset_t(["-lll", "example[4,95]", "example5"], None, b"@bar example[4-5,95] 3/100\n@foo example[4-5,95] 3/100\n@moo example[4-5,95] 3/100\n")
        # test empty result
        self._nodeset_t(["-l", "foo[3-70]", "bar6"], None, b"")
        # more arg-mixed tests
        self._nodeset_t(["-a", "-l"], None, b"@bar\n@foo\n@moo\n")
        self._nodeset_t(["-a", "-l", "-x example[1-100]"], None, b"")
        self._nodeset_t(["-a", "-l", "-x example[1-40]"], None, b"@bar\n@foo\n@moo\n")
        self._nodeset_t(["-l", "-x example3"], None, b"") # no -a, remove from nothing
        self._nodeset_t(["-l", "-i example3"], None, b"") # no -a, intersect from nothing
        self._nodeset_t(["-l", "-X example3"], None, b"@bar\n@foo\n@moo\n") # no -a, xor from nothing
        self._nodeset_t(["-l", "-", "-i example3"], "example[3,500]\n", b"@bar\n@foo\n@moo\n")

    def test_023_list_all(self):
        """test nodeset --list-all"""
        self._nodeset_t(["--list-all"], None, b"@bar\n@foo\n@moo\n")
        self._nodeset_t(["-L"], None, b"@bar\n@foo\n@moo\n")
        self._nodeset_t(["-LL"], None, b"@bar example[1-100]\n@foo example[1-100]\n@moo example[1-100]\n")
        self._nodeset_t(["-LLL"], None, b"@bar example[1-100] 100\n@foo example[1-100] 100\n@moo example[1-100] 100\n")


class CLINodesetGroupResolverTest2(CLINodesetTestBase):
    """Unit test class for testing CLI/Nodeset.py with custom Group Resolver"""

    def setUp(self):
        # Special tests that require a default group source set
        self.f = make_temp_file(dedent("""
            [Main]
            default: test

            [test]
            map: echo example[1-100]
            all: echo @foo,@bar,@moo
            list: echo foo bar moo

            [other]
            map: echo nova[030-489]
            all: echo @baz,@qux,@norf
            list: echo baz qux norf
            """).encode())
        set_std_group_resolver(GroupResolverConfig(self.f.name))

    def tearDown(self):
        set_std_group_resolver(None)
        self.f = None  # used to release temp file

    def test_024_groups(self):
        self._nodeset_t(["--split=2", "-r", "unknown2", "unknown3"], None, b"unknown2\nunknown3\n")
        self._nodeset_t(["-f", "-a"], None, b"example[1-100]\n")
        self._nodeset_t(["-f", "@moo"], None, b"example[1-100]\n")
        self._nodeset_t(["-f", "@moo", "@bar"], None, b"example[1-100]\n")
        self._nodeset_t(["-e", "-a"], None, ' '.join(["example%d" % i for i in range(1, 101)]).encode() + b'\n')
        self._nodeset_t(["-c", "-a"], None, b"100\n")
        self._nodeset_t(["-r", "-a"], None, b"@bar\n")
        self._nodeset_t(["--split=2", "-r", "unknown2", "unknown3"], None, b"unknown2\nunknown3\n")

    # We need to split following unit tests in order to reset group
    # source in setUp/tearDown...

    def test_025_groups(self):
        self._nodeset_t(["-s", "test", "-c", "-a", "-d"], None, b"100\n")

    def test_026_groups(self):
        self._nodeset_t(["-s", "test", "-r", "-a"], None, b"@test:bar\n")

    def test_027_groups(self):
        self._nodeset_t(["-s", "test", "-G", "-r", "-a"], None, b"@bar\n")

    def test_028_groups(self):
        self._nodeset_t(["-s", "test", "--groupsources"], None, b"test (default)\nother\n")

    def test_029_groups(self):
        self._nodeset_t(["-s", "test", "-q", "--groupsources"], None, b"test\nother\n")

    def test_030_groups(self):
        self._nodeset_t(["-f", "-a", "-"], "example101\n", b"example[1-101]\n")
        self._nodeset_t(["-f", "-a", "-"], "example102 example101\n", b"example[1-102]\n")

    # Check default group source switching...

    def test_031_groups(self):
        self._nodeset_t(["-s", "other", "-c", "-a", "-d"], None, b"460\n")
        self._nodeset_t(["-s", "test", "-c", "-a", "-d"], None, b"100\n")

    def test_032_groups(self):
        self._nodeset_t(["-s", "other", "-r", "-a"], None, b"@other:baz\n")
        self._nodeset_t(["-s", "test", "-r", "-a"], None, b"@test:bar\n")

    def test_033_groups(self):
        self._nodeset_t(["-s", "other", "-G", "-r", "-a"], None, b"@baz\n")
        self._nodeset_t(["-s", "test", "-G", "-r", "-a"], None, b"@bar\n")

    def test_034_groups(self):
        self._nodeset_t(["--groupsources"], None, b"test (default)\nother\n")

    def test_035_groups(self):
        self._nodeset_t(["-s", "other", "--groupsources"], None, b"other (default)\ntest\n")

    def test_036_groups(self):
        self._nodeset_t(["--groupsources"], None, b"test (default)\nother\n")

    def test_037_groups_output_format(self):
        self._nodeset_t(["-r", "-O", "{%s}", "-a"], None, b"{@bar}\n")

    def test_038_groups_output_format(self):
        self._nodeset_t(["-O", "{%s}", "-s", "other", "-r", "-a"], None, b"{@other:baz}\n")

    def test_039_list_all(self):
        """test nodeset --list-all (multi sources)"""
        self._nodeset_t(["--list-all"], None,
                        b"@bar\n@foo\n@moo\n@other:baz\n@other:norf\n@other:qux\n")
        self._nodeset_t(["--list-all", "-G"], None,
                        b"@bar\n@foo\n@moo\n@baz\n@norf\n@qux\n")
        self._nodeset_t(["-GL"], None,
                        b"@bar\n@foo\n@moo\n@baz\n@norf\n@qux\n")
        self._nodeset_t(["--list-all", "-s", "other"], None,
                        b"@other:baz\n@other:norf\n@other:qux\n@test:bar\n@test:foo\n@test:moo\n")
        self._nodeset_t(["--list-all", "-G", "-s", "other"], None,
                        b"@baz\n@norf\n@qux\n@bar\n@foo\n@moo\n") # 'other' source first

    def test_040_wildcards(self):
        """test nodeset with wildcards"""
        self._nodeset_t(["-f", "*"], None, b"example[1-100]\n")
        self._nodeset_t(["-f", "x*"], None, b"\n")  # no match
        self._nodeset_t(["-s", "other", "-f", "*"], None, b"nova[030-489]\n")
        self._nodeset_t(["-G", "-s", "other", "-f", "*"], None,
                        b"nova[030-489]\n")
        self._nodeset_t(["-s", "other", "-f", "nova0??"], None,
                        b"nova[030-099]\n")
        self._nodeset_t(["-s", "other", "-f", "nova?[12-42]"], None,
                        b"nova[030-042,112-142,212-242,312-342,412-442]\n")
        self._nodeset_t(["-s", "other", "-f", "*!*[033]"], None,
                        b"nova[030-032,034-489]\n")
        self._nodeset_t(["-s", "other", "--autostep=3", "-f", "*!*[033-099/2]"],
                        None, b"nova[030-031,032-100/2,101-489]\n")


class CLINodesetGroupResolverTest3(CLINodesetTestBase):
    """Unit test class for testing CLI/Nodeset.py with custom Group Resolver

    A case we support: one of the source misses the list upcall.
    """

    def setUp(self):
        # Special tests that require a default group source set
        self.f = make_temp_file(dedent("""
            [Main]
            default: test

            [test]
            map: echo example[1-100]
            all: echo @foo,@bar,@moo
            list: echo foo bar moo

            [other]
            map: echo nova[030-489]
            all: echo @baz,@qux,@norf
            list: echo baz qux norf

            [pdu]
            map: echo pdu-[0-3]-[1-2]
            """).encode())
        set_std_group_resolver(GroupResolverConfig(self.f.name))

    def tearDown(self):
        set_std_group_resolver(None)
        self.f = None  # used to release temp file

    def test_list_all(self):
        """test nodeset --list-all (w/ missing list upcall)"""
        self._nodeset_t(["--list-all"], None,
                        b"@bar\n@foo\n@moo\n@other:baz\n@other:norf\n@other:qux\n", 0,
                        b"Warning: No list upcall defined for group source pdu\n")
        self._nodeset_t(["-LL"], None,
                        b"@bar example[1-100]\n@foo example[1-100]\n@moo example[1-100]\n"
                        b"@other:baz nova[030-489]\n@other:norf nova[030-489]\n@other:qux nova[030-489]\n", 0,
                        b"Warning: No list upcall defined for group source pdu\n")
        self._nodeset_t(["-LLL"], None,
                        b"@bar example[1-100] 100\n@foo example[1-100] 100\n@moo example[1-100] 100\n"
                        b"@other:baz nova[030-489] 460\n@other:norf nova[030-489] 460\n@other:qux nova[030-489] 460\n", 0,
                        b"Warning: No list upcall defined for group source pdu\n")

    def test_list_failure(self):
        """test nodeset --list -s source w/ missing list upcall"""
        self._nodeset_t(["--list", "-s", "pdu"], None, b"", 1,
                        b'No list upcall defined for group source "pdu"\n')


class CLINodesetGroupResolverConfigErrorTest(CLINodesetTestBase):
    """Unit test class for testing GroupResolverConfigError"""

    def setUp(self):
        self.dname = make_temp_dir()
        self.gconff = make_temp_file(dedent("""
            [Main]
            default: default
            autodir: %s
            """ % self.dname).encode('ascii'))
        self.yamlf = make_temp_file(dedent("""
            default:
                compute: 'foo'
            broken: i am not a dict
            """).encode('ascii'), suffix=".yaml", dir=self.dname)

        set_std_group_resolver(GroupResolverConfig(self.gconff.name))

    def tearDown(self):
        set_std_group_resolver(None)
        self.gconff = None
        self.yamlf = None
        os.rmdir(self.dname)

    def test_bad_yaml_config(self):
        """test nodeset with bad yaml config"""
        self._nodeset_t(["--list-all"], None, b"", 1,
                        b"invalid content (group source 'broken' is not a dict)\n")


class CLINodesetEmptyGroupsConf(CLINodesetTestBase):
    """Unit test class for testing empty groups.conf"""

    def setUp(self):
        self.gconff = make_temp_file(b"")
        set_std_group_resolver(GroupResolverConfig(self.gconff.name))

    def tearDown(self):
        set_std_group_resolver(None)
        self.gconff = None

    def test_empty_groups_conf(self):
        """test nodeset with empty groups.conf"""
        self._nodeset_t(["--list-all"], None, b"")


class CLINodesetMalformedGroupsConf(CLINodesetTestBase):
    """Unit test class for testing malformed groups.conf"""

    def setUp(self):
        self.gconff = make_temp_file(b"[Main")
        set_std_group_resolver(GroupResolverConfig(self.gconff.name))

    def tearDown(self):
        set_std_group_resolver(None)
        self.gconff = None

    def test_malformed_groups_conf(self):
        """test nodeset with malformed groups.conf"""
        self._nodeset_t(["--list-all"], None, b"", 1, b"'[Main'\n")


class CLINodesetGroupsConfOption(CLINodesetTestBase):
    """Unit test class for testing --groupsconf option"""

    def setUp(self):
        self.gconff = make_temp_file(dedent("""
            [Main]
            default: global_default

            [global_default]
            map: echo example[1-100]
            all: echo @foo,@bar,@moo
            list: echo foo bar moo
            """).encode())
        set_std_group_resolver_config(self.gconff.name)

        # passed to --groupsconf
        self.custf = make_temp_file(dedent("""
            [Main]
            default: custom

            [custom]
            map: echo custom[7-42]
            all: echo @selene,@artemis
            list: echo selene artemis
            """).encode())

    def tearDown(self):
        set_std_group_resolver(None)
        self.gconff = None
        self.custf = None

    def test_groupsconf_option(self):
        """test nodeset with --groupsconf"""
        self._nodeset_t(["--list-all"], None, b"@bar\n@foo\n@moo\n")
        self._nodeset_t(["-f", "@foo"], None, b"example[1-100]\n")
        self._nodeset_t(["--groupsconf", self.custf.name, "--list-all"], None, b"@artemis\n@selene\n")
        self._nodeset_t(["--groupsconf", self.custf.name, "-f", "@artemis"], None, b"custom[7-42]\n")



# ClusterShell (distant) test suite
# Written by S. Thiell

"""Unit test for ClusterShell Task (distant)"""

import pwd
import shutil
import warnings

from TLib import HOSTNAME, make_temp_filename, make_temp_dir
from ClusterShell.Event import EventHandler
from ClusterShell.Task import *
from ClusterShell.Worker.Ssh import WorkerSsh
from ClusterShell.Worker.EngineClient import *
from ClusterShell.Worker.Worker import FANOUT_UNLIMITED, WorkerBadArgumentError

import socket

# TEventHandlerChecker 'received event' flags
EV_START = 0x01
EV_PICKUP = 0x02
EV_READ = 0x04
EV_WRITTEN = 0x08
EV_HUP = 0x10
EV_TIMEOUT = 0x20
EV_CLOSE = 0x40


class TaskDistantMixin(object):

    def setUp(self):
        self._task = task_self()

    def testLocalhostCommand(self):
        # init worker
        worker = self._task.shell("/bin/hostname", nodes=HOSTNAME)
        # run task
        self._task.resume()

    def testLocalhostCommand2(self):
        # init worker
        worker = self._task.shell("/bin/hostname", nodes=HOSTNAME)

        worker = self._task.shell("/bin/uname -r", nodes=HOSTNAME)
        # run task
        self._task.resume()

    def testTaskShellWorkerGetCommand(self):
        worker1 = self._task.shell("/bin/hostname", nodes=HOSTNAME)
        worker2 = self._task.shell("/bin/uname -r", nodes=HOSTNAME)
        self._task.resume()
        self.assertTrue(hasattr(worker1, 'command'))
        self.assertTrue(hasattr(worker2, 'command'))
        self.assertEqual(worker1.command, "/bin/hostname")
        self.assertEqual(worker2.command, "/bin/uname -r")

    def testTaskShellRunDistant(self):
        wrk = task_self().run("false", nodes=HOSTNAME)
        self.assertEqual(wrk.node_retcode(HOSTNAME), 1)

    def testLocalhostCopy(self):
        dests = []
        try:
            for i in range(5):
                dest = make_temp_filename(suffix='LocalhostCopy')
                dests.append(dest)
                worker = self._task.copy("/etc/hosts", dest, nodes=HOSTNAME)
            self._task.resume()
        finally:
            for dest in dests:
                os.unlink(dest)

    def testCopyNodeFailure(self):
        # == stderr merged ==
        self._task.set_default("stderr", False)
        dest = make_temp_filename(suffix='LocalhostCopyF')
        worker = self._task.copy("/etc/hosts", dest,
                                 nodes='unlikely-node,%s' % HOSTNAME)
        self._task.resume()
        self.assertEqual(worker.node_error_buffer("unlikely-node"), None)
        self.assertTrue(len(worker.node_buffer("unlikely-node")) > 2)
        os.unlink(dest)

        # == stderr separated ==
        self._task.set_default("stderr", True)
        try:
            dest = make_temp_filename(suffix='LocalhostCopyF2')
            worker = self._task.copy("/etc/hosts", dest,
                                     nodes='unlikely-node,%s' % HOSTNAME)
            # run task
            self._task.resume()
            self.assertTrue(worker.node_buffer("unlikely-node") is None)
            self.assertTrue(len(worker.node_error_buffer("unlikely-node")) > 2)
            os.unlink(dest)
        finally:
            self._task.set_default("stderr", False)

    def testLocalhostCopyDir(self):
        dtmp_src = make_temp_dir('src')
        dtmp_dst = make_temp_dir('testLocalhostCopyDir')
        try:
            os.mkdir(os.path.join(dtmp_src, "lev1_a"))
            os.mkdir(os.path.join(dtmp_src, "lev1_b"))
            os.mkdir(os.path.join(dtmp_src, "lev1_a", "lev2"))
            worker = self._task.copy(dtmp_src, dtmp_dst, nodes=HOSTNAME)
            self._task.resume()
            self.assertTrue(os.path.exists(os.path.join(dtmp_dst,
                            os.path.basename(dtmp_src), "lev1_a", "lev2")))
        finally:
            shutil.rmtree(dtmp_dst, ignore_errors=True)
            shutil.rmtree(dtmp_src, ignore_errors=True)

    def testLocalhostExplicitSshCopy(self):
        dest = make_temp_filename('testLocalhostExplicitSshCopy')
        srcsz = os.path.getsize("/etc/hosts")
        try:
            worker = WorkerSsh(HOSTNAME, source="/etc/hosts", dest=dest,
                               handler=None, timeout=10)
            self._task.schedule(worker)
            self._task.resume()
            self.assertEqual(srcsz, os.path.getsize(dest))
        finally:
            os.remove(dest)

    def testLocalhostExplicitSshCopyWithOptions(self):
        dest = make_temp_dir('testLocalhostExplicitSshCopyWithOptions')
        self._task.set_info("scp_path", "/usr/bin/scp -l 10")
        self._task.set_info("scp_options", "-oLogLevel=QUIET")
        try:
            worker = WorkerSsh(HOSTNAME, source="/etc/hosts", dest=dest,
                               handler=None)
            self._task.schedule(worker)
            self._task.resume()
            self.assertEqual(self._task.max_retcode(), 0)
            self.assertTrue(os.path.exists(os.path.join(dest, "hosts")))
        finally:
            os.unlink(os.path.join(dest, "hosts"))
            os.rmdir(dest)
        # clear options after test
        task_cleanup()
        self.assertEqual(task_self().info("scp_path"), None)

    def testLocalhostExplicitSshCopyDir(self):
        dtmp_src = make_temp_dir('src')
        dtmp_dst = make_temp_dir('testLocalhostExplicitSshCopyDir')
        try:
            os.mkdir(os.path.join(dtmp_src, "lev1_a"))
            os.mkdir(os.path.join(dtmp_src, "lev1_b"))
            os.mkdir(os.path.join(dtmp_src, "lev1_a", "lev2"))
            worker = WorkerSsh(HOSTNAME, source=dtmp_src, dest=dtmp_dst,
                               handler=None)
            self._task.schedule(worker)
            self._task.resume()
            path = os.path.join(dtmp_dst, os.path.basename(dtmp_src), "lev1_a",
                                "lev2")
            self.assertTrue(os.path.exists(path))
        finally:
            shutil.rmtree(dtmp_dst, ignore_errors=True)
            shutil.rmtree(dtmp_src, ignore_errors=True)

    def testLocalhostExplicitSshCopyDirPreserve(self):
        dtmp_src = make_temp_dir('src')
        dtmp_dst = make_temp_dir('testLocalhostExplicitSshCopyDirPreserve')
        try:
            os.mkdir(os.path.join(dtmp_src, "lev1_a"))
            os.mkdir(os.path.join(dtmp_src, "lev1_b"))
            os.mkdir(os.path.join(dtmp_src, "lev1_a", "lev2"))
            worker = WorkerSsh(HOSTNAME, source=dtmp_src, dest=dtmp_dst,
                               handler=None, timeout=10, preserve=True)
            self._task.schedule(worker)
            self._task.resume()
            self.assertTrue(os.path.exists(os.path.join(dtmp_dst,
                            os.path.basename(dtmp_src), "lev1_a", "lev2")))
        finally:
            shutil.rmtree(dtmp_dst, ignore_errors=True)
            shutil.rmtree(dtmp_src, ignore_errors=True)

    def testExplicitSshWorker(self):
        # init worker
        worker = WorkerSsh(HOSTNAME, command="/bin/echo alright", handler=None)
        self._task.schedule(worker)
        # run task
        self._task.resume()
        # test output
        self.assertEqual(worker.node_buffer(HOSTNAME), b"alright")

    def testExplicitSshWorkerWithOptions(self):
        self._task.set_info("ssh_path", "/usr/bin/ssh -C")
        self._task.set_info("ssh_options", "-oLogLevel=QUIET")
        worker = WorkerSsh(HOSTNAME, command="/bin/echo alright", handler=None)
        self._task.schedule(worker)
        # run task
        self._task.resume()
        # test output
        self.assertEqual(worker.node_buffer(HOSTNAME), b"alright")
        # clear options after test
        task_cleanup()
        self.assertEqual(task_self().info("ssh_path"), None)

    def testExplicitSshWorkerStdErr(self):
        # init worker
        worker = WorkerSsh(HOSTNAME, command="/bin/echo alright 1>&2",
                           handler=None, stderr=True)
        self._task.schedule(worker)
        # run task
        self._task.resume()
        # test output
        self.assertEqual(worker.node_error_buffer(HOSTNAME), b"alright")

        # Re-test with stderr=False
        worker = WorkerSsh(HOSTNAME, command="/bin/echo alright 1>&2",
                           handler=None, stderr=False)
        self._task.schedule(worker)
        # run task
        self._task.resume()
        # test output
        self.assertEqual(worker.node_error_buffer(HOSTNAME), None)

    class TEventHandlerChecker(EventHandler):
        """simple event trigger validator"""
        def __init__(self, test):
            self.test = test
            self.flags = 0
            self.read_count = 0
            self.written_count = 0
        def ev_start(self, worker):
            self.test.assertEqual(self.flags, 0)
            self.flags |= EV_START
        def ev_pickup(self, worker):
            self.test.assertTrue(self.flags & EV_START)
            self.flags |= EV_PICKUP
            self.last_node = worker.current_node
        def ev_read(self, worker):
            self.test.assertEqual(self.flags, EV_START | EV_PICKUP)
            self.flags |= EV_READ
            self.last_node = worker.current_node
            self.last_read = worker.current_msg
        def ev_written(self, worker):
            self.test.assertTrue(self.flags & (EV_START | EV_PICKUP))
            self.flags |= EV_WRITTEN
        def ev_hup(self, worker):
            self.test.assertTrue(self.flags & (EV_START | EV_PICKUP))
            self.flags |= EV_HUP
            self.last_node = worker.current_node
            self.last_rc = worker.current_rc
        def ev_timeout(self, worker):
            self.test.assertTrue(self.flags & EV_START)
            self.flags |= EV_TIMEOUT
            self.last_node = worker.current_node
        def ev_close(self, worker):
            self.test.assertTrue(self.flags & EV_START)
            self.test.assertTrue(self.flags & EV_CLOSE == 0)
            self.flags |= EV_CLOSE

    def testShellEvents(self):
        # init worker
        test_eh = self.__class__.TEventHandlerChecker(self)
        worker = self._task.shell("/bin/hostname", nodes=HOSTNAME, handler=test_eh)
        # run task
        self._task.resume()
        # test events received: start, read, hup, close
        self.assertEqual(test_eh.flags, EV_START | EV_PICKUP | EV_READ | EV_HUP | EV_CLOSE)

    def testShellEventsWithTimeout(self):
        # init worker
        test_eh = self.__class__.TEventHandlerChecker(self)
        worker = self._task.shell("/bin/echo alright && /bin/sleep 10",
                                  nodes=HOSTNAME, handler=test_eh, timeout=2)
        self.assertTrue(worker != None)
        # run task
        self._task.resume()
        # test events received: start, read, timeout, close
        self.assertEqual(test_eh.flags, EV_START | EV_PICKUP | EV_READ | EV_TIMEOUT | EV_CLOSE)
        self.assertEqual(worker.node_buffer(HOSTNAME), b"alright")
        self.assertEqual(worker.num_timeout(), 1)
        self.assertEqual(self._task.num_timeout(), 1)
        count = 0
        for node in self._task.iter_keys_timeout():
            count += 1
            self.assertEqual(node, HOSTNAME)
        self.assertEqual(count, 1)
        count = 0
        for node in worker.iter_keys_timeout():
            count += 1
            self.assertEqual(node, HOSTNAME)
        self.assertEqual(count, 1)

    def testShellEventsWithTimeout2(self):
        # init worker
        test_eh1 = self.__class__.TEventHandlerChecker(self)
        worker1 = self._task.shell("/bin/echo alright && /bin/sleep 10",
                                   nodes=HOSTNAME, handler=test_eh1, timeout=2)
        test_eh2 = self.__class__.TEventHandlerChecker(self)
        worker2 = self._task.shell("/bin/echo okay && /bin/sleep 10",
                                   nodes=HOSTNAME, handler=test_eh2, timeout=3)
        # run task
        self._task.resume()
        # test events received: start, read, timeout, close
        self.assertEqual(test_eh1.flags, EV_START | EV_PICKUP | EV_READ | EV_TIMEOUT | EV_CLOSE)
        self.assertEqual(test_eh2.flags, EV_START | EV_PICKUP | EV_READ | EV_TIMEOUT | EV_CLOSE)
        self.assertEqual(worker1.node_buffer(HOSTNAME), b"alright")
        self.assertEqual(worker2.node_buffer(HOSTNAME), b"okay")
        self.assertEqual(worker1.num_timeout(), 1)
        self.assertEqual(worker2.num_timeout(), 1)
        self.assertEqual(self._task.num_timeout(), 2)

    def testShellEventsReadNoEOL(self):
        # init worker
        test_eh = self.__class__.TEventHandlerChecker(self)
        worker = self._task.shell("/bin/echo -n okay", nodes=HOSTNAME, handler=test_eh)
        # run task
        self._task.resume()
        # test events received: start, close
        self.assertEqual(test_eh.flags, EV_START | EV_PICKUP | EV_READ | EV_HUP | EV_CLOSE)
        self.assertEqual(worker.node_buffer(HOSTNAME), b"okay")

    def testShellEventsNoReadNoTimeout(self):
        # init worker
        test_eh = self.__class__.TEventHandlerChecker(self)
        worker = self._task.shell("/bin/sleep 2", nodes=HOSTNAME, handler=test_eh)
        # run task
        self._task.resume()
        # test events received: start, close
        self.assertEqual(test_eh.flags, EV_START | EV_PICKUP | EV_HUP | EV_CLOSE)
        self.assertEqual(worker.node_buffer(HOSTNAME), None)

    def testLocalhostCommandFanout(self):
        fanout = self._task.info("fanout")
        self._task.set_info("fanout", 2)
        # init worker
        for i in range(0, 10):
            worker = self._task.shell("/bin/echo %d" % i, nodes=HOSTNAME)
        # run task
        self._task.resume()
        # restore fanout value
        self._task.set_info("fanout", fanout)

    def testWorkerBuffers(self):
        # Warning: if you modify this test, please also modify testWorkerErrorBuffers()
        task = task_self()
        worker = task.shell("/usr/bin/printf 'foo\nbar\nxxx\n'",
                            nodes=HOSTNAME)
        task.resume()
        # test iter_buffers() by worker...
        cnt = 2
        for buf, nodes in worker.iter_buffers():
            cnt -= 1
            if buf == b"foo\nbar\nxxx\n":
                self.assertEqual(len(nodes), 1)
                self.assertEqual(str(nodes), HOSTNAME)
        self.assertEqual(cnt, 1)
        # new check in 1.7 to ensure match_keys is not a string
        testgen = worker.iter_buffers(HOSTNAME)
        # cast to list to effectively iterate
        self.assertRaises(TypeError, list, testgen)
        # and also fixed an issue when match_keys was an empty list
        for buf, nodes in worker.iter_buffers([]):
            self.assertFalse("Found buffer with empty match_keys?!")
        for buf, nodes in worker.iter_buffers([HOSTNAME]):
            cnt -= 1
            if buf == b"foo\nbar\nxxx\n":
                self.assertEqual(len(nodes), 1)
                self.assertEqual(str(nodes), HOSTNAME)
        self.assertEqual(cnt, 0)
        # test flushing buffers by worker
        worker.flush_buffers()
        self.assertEqual(list(worker.iter_buffers()), [])

    def testWorkerErrorBuffers(self):
        # Warning: if you modify this test, please also modify testWorkerBuffers()
        task = task_self()
        worker = task.shell("/usr/bin/printf 'foo\nbar\nxxx\n' 1>&2",
                            nodes=HOSTNAME, stderr=True)
        task.resume()
        # test iter_errors() by worker...
        cnt = 2
        for buf, nodes in worker.iter_errors():
            cnt -= 1
            if buf == b"foo\nbar\nxxx\n":
                self.assertEqual(len(nodes), 1)
                self.assertEqual(str(nodes), HOSTNAME)
        self.assertEqual(cnt, 1)
        # new check in 1.7 to ensure match_keys is not a string
        testgen = worker.iter_errors(HOSTNAME)
        # cast to list to effectively iterate
        self.assertRaises(TypeError, list, testgen)
        # and also fixed an issue when match_keys was an empty list
        for buf, nodes in worker.iter_errors([]):
            self.assertFalse("Found error buffer with empty match_keys?!")
        for buf, nodes in worker.iter_errors([HOSTNAME]):
            cnt -= 1
            if buf == b"foo\nbar\nxxx\n":
                self.assertEqual(len(nodes), 1)
                self.assertEqual(str(nodes), HOSTNAME)
        self.assertEqual(cnt, 0)
        # test flushing error buffers by worker
        worker.flush_errors()
        self.assertEqual(list(worker.iter_errors()), [])

    def testWorkerNodeBuffers(self):
        task = task_self()

        worker = task.shell("/usr/bin/printf 'foo\nbar\nxxx\n'",
                            nodes=HOSTNAME)

        task.resume()

        cnt = 1
        for node, buf in worker.iter_node_buffers():
            cnt -= 1
            if buf == b"foo\nbar\nxxx\n":
                self.assertEqual(node, HOSTNAME)
        self.assertEqual(cnt, 0)

    def testWorkerNodeErrors(self):
        task = task_self()

        worker = task.shell("/usr/bin/printf 'foo\nbar\nxxx\n' 1>&2",
                            nodes=HOSTNAME, stderr=True)

        task.resume()

        cnt = 1
        for node, buf in worker.iter_node_errors():
            cnt -= 1
            if buf == b"foo\nbar\nxxx\n":
                self.assertEqual(node, HOSTNAME)
        self.assertEqual(cnt, 0)

    def testWorkerRetcodes(self):
        task = task_self()

        worker = task.shell("/bin/sh -c 'exit 3'", nodes=HOSTNAME)

        task.resume()

        cnt = 2
        for rc, keys in worker.iter_retcodes():
            cnt -= 1
            self.assertEqual(rc, 3)
            self.assertEqual(len(keys), 1)
            self.assertEqual(keys[0], HOSTNAME)

        self.assertEqual(cnt, 1)

        for rc, keys in worker.iter_retcodes(HOSTNAME):
            cnt -= 1
            self.assertEqual(rc, 3)
            self.assertEqual(len(keys), 1)
            self.assertEqual(keys[0], HOSTNAME)

        self.assertEqual(cnt, 0)

        # test node_retcode
        self.assertEqual(worker.node_retcode(HOSTNAME), 3)   # 1.2.91+
        self.assertEqual(worker.node_rc(HOSTNAME), 3)

        # test node_retcode failure
        self.assertRaises(KeyError, worker.node_retcode, "dummy")

        # test max retcode API
        self.assertEqual(task.max_retcode(), 3)

    def testWorkerNodeRetcodes(self):
        task = task_self()

        worker = task.shell("/bin/sh -c 'exit 3'", nodes=HOSTNAME)

        task.resume()

        cnt = 1
        for node, rc in worker.iter_node_retcodes():
            cnt -= 1
            self.assertEqual(rc, 3)
            self.assertEqual(node, HOSTNAME)

        self.assertEqual(cnt, 0)

    def testEscape(self):
        cmd = r"export CSTEST=foobar; /bin/echo \$CSTEST | sed 's/\ foo/bar/'"
        worker = self._task.shell(cmd, nodes=HOSTNAME)
        # execute
        self._task.resume()
        # read result
        self.assertEqual(worker.node_buffer(HOSTNAME), b"$CSTEST")

    def testEscape2(self):
        cmd = r"export CSTEST=foobar; /bin/echo $CSTEST | sed 's/\ foo/bar/'"
        worker = self._task.shell(cmd, nodes=HOSTNAME)
        # execute
        self._task.resume()
        # read result
        self.assertEqual(worker.node_buffer(HOSTNAME), b"foobar")

    def testSshUserOption(self):
        ssh_user_orig = self._task.info("ssh_user")
        self._task.set_info("ssh_user", pwd.getpwuid(os.getuid())[0])
        worker = self._task.shell("/bin/echo foobar", nodes=HOSTNAME)
        self._task.resume()
        # restore original ssh_user (None)
        self.assertEqual(ssh_user_orig, None)
        self._task.set_info("ssh_user", ssh_user_orig)

    def testSshUserOptionForScp(self):
        ssh_user_orig = self._task.info("ssh_user")
        self._task.set_info("ssh_user", pwd.getpwuid(os.getuid())[0])
        dest = make_temp_filename('testLocalhostCopyU')
        worker = self._task.copy("/etc/hosts", dest, nodes=HOSTNAME)
        self._task.resume()
        # restore original ssh_user (None)
        self.assertEqual(ssh_user_orig, None)
        self._task.set_info("ssh_user", ssh_user_orig)
        os.unlink(dest)

    def testSshOptionsOption(self):
        ssh_options_orig = self._task.info("ssh_options")
        try:
            self._task.set_info("ssh_options", "-oLogLevel=QUIET")
            worker = self._task.shell("/bin/echo foobar", nodes=HOSTNAME)
            self._task.resume()
            self.assertEqual(worker.node_buffer(HOSTNAME), b"foobar")
            # test 3 options
            self._task.set_info("ssh_options", \
                "-oLogLevel=QUIET -oStrictHostKeyChecking=no -oVerifyHostKeyDNS=no")
            worker = self._task.shell("/bin/echo foobar3", nodes=HOSTNAME)
            self._task.resume()
            self.assertEqual(worker.node_buffer(HOSTNAME), b"foobar3")
        finally:
            # restore original ssh_user (None)
            self.assertEqual(ssh_options_orig, None)
            self._task.set_info("ssh_options", ssh_options_orig)

    def testSshOptionsOptionForScp(self):
        ssh_options_orig = self._task.info("ssh_options")
        testfile = None
        try:
            testfile = make_temp_filename('testLocalhostCopyO')
            if os.path.exists(testfile):
                os.remove(testfile)
            self._task.set_info("ssh_options", \
                "-oLogLevel=QUIET -oStrictHostKeyChecking=no -oVerifyHostKeyDNS=no")
            worker = self._task.copy("/etc/hosts", testfile, nodes=HOSTNAME)
            self._task.resume()
            self.assertTrue(os.path.exists(testfile))
        finally:
            os.unlink(testfile)
            # restore original ssh_user (None)
            self.assertEqual(ssh_options_orig, None)
            self._task.set_info("ssh_options", ssh_options_orig)

    def testShellStderrWithHandler(self):
        class StdErrHandler(EventHandler):
            def ev_error(self, worker):
                assert worker.current_errmsg == b"something wrong"

        worker = self._task.shell("echo something wrong 1>&2", nodes=HOSTNAME,
                                  handler=StdErrHandler(), stderr=True)
        self._task.resume()
        for buf, nodes in worker.iter_errors():
            self.assertEqual(buf, b"something wrong")
        for buf, nodes in worker.iter_errors([HOSTNAME]):
            self.assertEqual(buf, b"something wrong")

    def testShellWriteSimple(self):
        worker = self._task.shell("cat", nodes=HOSTNAME)
        worker.write(b"this is a test\n")
        worker.set_write_eof()
        self._task.resume()
        self.assertEqual(worker.node_buffer(HOSTNAME), b"this is a test")

    def testShellWriteHandler(self):
        class WriteOnReadHandler(EventHandler):
            def __init__(self, target_worker):
                self.target_worker = target_worker
            def ev_read(self, worker):
                self.target_worker.write(worker.current_node.encode('utf-8')
                                         + b':' + worker.current_msg + b'\n')
                self.target_worker.set_write_eof()

        reader = self._task.shell("cat", nodes=HOSTNAME)
        worker = self._task.shell("sleep 1; echo foobar", nodes=HOSTNAME,
                                  handler=WriteOnReadHandler(reader))
        self._task.resume()
        res = "%s:foobar" % HOSTNAME
        self.assertEqual(reader.node_buffer(HOSTNAME), res.encode('utf-8'))

    def testSshBadArgumentOption(self):
        # Check code < 1.4 compatibility
        self.assertRaises(WorkerBadArgumentError, WorkerSsh, HOSTNAME, None,
                          None)
        # As of 1.4, ValueError is raised for missing parameter
        self.assertRaises(ValueError, WorkerSsh, HOSTNAME, None, None) # 1.4+

    def testCopyEvents(self):
        test_eh = self.__class__.TEventHandlerChecker(self)
        dest = make_temp_filename('testLocalhostCopyEvents')
        worker = self._task.copy("/etc/hosts", dest, nodes=HOSTNAME,
                                 handler=test_eh)
        # run task
        self._task.resume()
        os.unlink(dest)
        self.assertEqual(test_eh.flags, EV_START | EV_PICKUP | EV_HUP | EV_CLOSE)

    def testWorkerAbort(self):
        task = task_self()

        # Test worker.abort() in an event handler.
        class AbortOnTimer(EventHandler):
            def __init__(self, worker):
                EventHandler.__init__(self)
                self.ext_worker = worker
                self.testtimer = False
            def ev_timer(self, timer):
                self.ext_worker.abort()
                self.ext_worker.abort()  # safe but no effect
                self.testtimer = True

        aot = AbortOnTimer(task.shell("sleep 10", nodes=HOSTNAME))
        self.assertEqual(aot.testtimer, False)
        task.timer(1.5, handler=aot)
        task.resume()
        self.assertEqual(aot.testtimer, True)

    def testWorkerAbortSanity(self):
        task = task_self()
        worker = task.shell("sleep 1", nodes=HOSTNAME)
        worker.abort()

        # test noop abort() on unscheduled worker
        worker = WorkerSsh(HOSTNAME, command="sleep 1", handler=None,
                           timeout=None)
        worker.abort()

    def testLocalhostRCopy(self):
        try:
            dest = make_temp_dir('testLocalhostRCopy')
            # use fake node 'aaa' to test rank > 0
            worker = self._task.rcopy("/etc/hosts", dest, "aaa,%s" % HOSTNAME,
                                      handler=None, timeout=10)
            self._task.resume()
            self.assertEqual(worker.source, "/etc/hosts")
            self.assertEqual(worker.dest, dest)
            self.assertTrue(os.path.exists(os.path.join(dest, "hosts.%s" % HOSTNAME)))
        finally:
            shutil.rmtree(dest, ignore_errors=True)

    def testLocalhostExplicitSshReverseCopy(self):
        dest = make_temp_dir('testLocalhostExplicitSshRCopy')
        try:
            worker = WorkerSsh(HOSTNAME, source="/etc/hosts", dest=dest,
                               handler=None, timeout=10, reverse=True)
            self._task.schedule(worker)
            self._task.resume()
            self.assertEqual(worker.source, "/etc/hosts")
            self.assertEqual(worker.dest, dest)
            self.assertTrue(os.path.exists(os.path.join(dest, "hosts.%s" % HOSTNAME)))
        finally:
            shutil.rmtree(dest, ignore_errors=True)

    def testLocalhostExplicitSshReverseCopyDir(self):
        dtmp_src = make_temp_dir('src')
        dtmp_dst = make_temp_dir('testLocalhostExplicitSshReverseCopyDir')
        try:
            os.mkdir(os.path.join(dtmp_src, "lev1_a"))
            os.mkdir(os.path.join(dtmp_src, "lev1_b"))
            os.mkdir(os.path.join(dtmp_src, "lev1_a", "lev2"))
            worker = WorkerSsh(HOSTNAME, source=dtmp_src, dest=dtmp_dst,
                               handler=None, timeout=30, reverse=True)
            self._task.schedule(worker)
            self._task.resume()
            self.assertTrue(os.path.exists(os.path.join(dtmp_dst,
                "%s.%s" % (os.path.basename(dtmp_src), HOSTNAME), "lev1_a", "lev2")))
        finally:
            shutil.rmtree(dtmp_dst, ignore_errors=True)
            shutil.rmtree(dtmp_src, ignore_errors=True)

    def testLocalhostExplicitSshReverseCopyDirPreserve(self):
        dtmp_src = make_temp_dir('src')
        dtmp_dst = make_temp_dir('testLocalhostExplicitSshReverseCopyDirPreserve')
        try:
            os.mkdir(os.path.join(dtmp_src, "lev1_a"))
            os.mkdir(os.path.join(dtmp_src, "lev1_b"))
            os.mkdir(os.path.join(dtmp_src, "lev1_a", "lev2"))
            worker = WorkerSsh(HOSTNAME, source=dtmp_src, dest=dtmp_dst,
                               handler=None, timeout=30, reverse=True)
            self._task.schedule(worker)
            self._task.resume()
            self.assertTrue(os.path.exists(os.path.join(dtmp_dst,
                "%s.%s" % (os.path.basename(dtmp_src), HOSTNAME), "lev1_a", "lev2")))
        finally:
            shutil.rmtree(dtmp_dst, ignore_errors=True)
            shutil.rmtree(dtmp_src, ignore_errors=True)

    def testErroneousSshPath(self):
        try:
            self._task.set_info("ssh_path", "/wrong/path/to/ssh")
            # init worker
            worker = self._task.shell("/bin/echo ok", nodes=HOSTNAME)
            # run task
            self._task.resume()
            self.assertEqual(self._task.max_retcode(), 255)
        finally:
            # restore fanout value
            self._task.set_info("ssh_path", None)

    class TEventHandlerEvCountChecker(EventHandler):
        """simple event count validator"""

        def __init__(self):
            self.start_count = 0
            self.pickup_count = 0
            self.hup_count = 0
            self.close_count = 0

        def ev_start(self, worker):
            self.start_count += 1

        def ev_pickup(self, worker):
            self.pickup_count += 1

        def ev_hup(self, worker):
            self.hup_count += 1

        def ev_close(self, worker):
            self.close_count += 1

    def testWorkerEventCount(self):
        test_eh = self.__class__.TEventHandlerEvCountChecker()
        nodes = "localhost,%s" % HOSTNAME
        worker = self._task.shell("/bin/hostname", nodes=nodes, handler=test_eh)
        self._task.resume()
        # test event count
        self.assertEqual(test_eh.pickup_count, 2)
        self.assertEqual(test_eh.hup_count, 2)
        self.assertEqual(test_eh.start_count, 1)
        self.assertEqual(test_eh.close_count, 1)

    def test_last_deprecated(self):

        class TestHandlerHandler(EventHandler):
            def ev_read(self, worker):
                with warnings.catch_warnings(record=True) as wngs:
                    warnings.simplefilter("always")
                    self.node, self.msg = worker.last_read()
                    assert len(wngs) == 1
                    assert issubclass(wngs[-1].category, DeprecationWarning)
            def ev_hup(self, worker):
                with warnings.catch_warnings(record=True) as wngs:
                    warnings.simplefilter("always")
                    self.node, self.rc = worker.last_retcode()
                    assert len(wngs) == 1
                    assert issubclass(wngs[-1].category, DeprecationWarning)

        eh = TestHandlerHandler()
        reader = self._task.shell("echo foobar", nodes=HOSTNAME, handler=eh)
        self._task.resume()
        self.assertEqual(eh.node, HOSTNAME)
        self.assertEqual(eh.rc, 0)



# ClusterShell.Topology test suite
# Written by H. Doreau

"""Unit test for Topology"""

import unittest
import tempfile

# profiling imports
#import cProfile
#from guppy import hpy
# ---

from ClusterShell.Topology import *
from ClusterShell.NodeSet import NodeSet


class TopologyTest(unittest.TestCase):

    def testInvalidConfigurationFile(self):
        """test detecting invalid configuration file"""
        parser = TopologyParser()
        self.assertRaises(TopologyError,
                          parser.load,
                          '/invalid/path/for/testing')
        self.assertRaises(TopologyError,
                          TopologyParser,
                          '/invalid/path/for/testing')

    def testTopologyGraphGeneration(self):
        """test graph generation"""
        g = TopologyGraph()
        ns1 = NodeSet('nodes[0-5]')
        ns2 = NodeSet('nodes[6-10]')
        g.add_route(ns1, ns2)
        self.assertEqual(g.dest(ns1), ns2)

    def testAddingSeveralRoutes(self):
        """test adding several valid routes"""
        g = TopologyGraph()
        admin = NodeSet('admin')
        ns0 = NodeSet('nodes[0-9]')
        ns1 = NodeSet('nodes[10-19]')
        g.add_route(admin, ns0)
        g.add_route(ns0, ns1)
        # Connect a new dst nodeset to an existing src
        ns2 = NodeSet('nodes[20-29]')
        g.add_route(ns0, ns2)
        # Add the same dst nodeset twice (no error)
        g.add_route(ns0, ns2)

        self.assertEquals(g.dest(admin), ns0)
        self.assertEquals(g.dest(ns0), ns1 | ns2)

    def testBadLink(self):
        """test detecting bad links in graph"""
        g = TopologyGraph()
        admin = NodeSet('admin')
        ns0 = NodeSet('nodes[0-9]')
        ns1 = NodeSet('nodes[10-19]')
        g.add_route(admin, ns0)
        g.add_route(ns0, ns1)
        # Add a known src nodeset as a dst nodeset (error!)
        self.assertRaises(TopologyError, g.add_route, ns1, ns0)

    def testOverlappingRoutes(self):
        """test overlapping routes detection"""
        g = TopologyGraph()
        admin = NodeSet('admin')
        # Add the same nodeset twice
        ns0 = NodeSet('nodes[0-9]')
        ns1 = NodeSet('nodes[10-19]')
        ns1_overlap = NodeSet('nodes[5-29]')

        self.assertRaises(TopologyError, g.add_route, ns0, ns0)
        g.add_route(ns0, ns1)
        self.assertRaises(TopologyError, g.add_route, ns0, ns1_overlap)

    def testBadTopologies(self):
        """test detecting invalid topologies"""
        g = TopologyGraph()
        admin = NodeSet('admin')
        # Add the same nodeset twice
        ns0 = NodeSet('nodes[0-9]')
        ns1 = NodeSet('nodes[10-19]')
        ns2 = NodeSet('nodes[20-29]')

        g.add_route(admin, ns0)
        g.add_route(ns0, ns1)
        g.add_route(ns0, ns2)

        # add a superset of a known destination as source
        ns2_sup = NodeSet('somenode[0-10]')
        ns2_sup.add(ns2)
        self.assertRaises(TopologyError, g.add_route, ns2_sup, NodeSet('foo1'))

        # Add a known dst nodeset as a src nodeset
        ns3 = NodeSet('nodes[30-39]')
        g.add_route(ns1, ns3)

        # Add a subset of a known src nodeset as src
        ns0_sub = NodeSet(','.join(ns0[:3:]))
        ns4 = NodeSet('nodes[40-49]')
        g.add_route(ns0_sub, ns4)

        # Add a subset of a known dst nodeset as src
        ns1_sub = NodeSet(','.join(ns1[:3:]))
        self.assertRaises(TopologyError, g.add_route, ns4, ns1_sub)
        # Add a subset of a known src nodeset as dst
        self.assertRaises(TopologyError, g.add_route, ns4, ns0_sub)
        # Add a subset of a known dst nodeset as dst
        self.assertRaises(TopologyError, g.add_route, ns4, ns1_sub)
        # src <- subset of -> dst
        ns5 = NodeSet('nodes[50-59]')
        ns5_sub = NodeSet(','.join(ns5[:3:]))
        self.assertRaises(TopologyError, g.add_route, ns5, ns5_sub)
        self.assertRaises(TopologyError, g.add_route, ns5_sub, ns5)

        self.assertEqual(g.dest(ns0), (ns1 | ns2))
        self.assertEqual(g.dest(ns1), ns3)
        self.assertEqual(g.dest(ns2), None)
        self.assertEqual(g.dest(ns3), None)
        self.assertEqual(g.dest(ns4), None)
        self.assertEqual(g.dest(ns5), None)
        self.assertEqual(g.dest(ns0_sub), (ns1 | ns2 | ns4))

        g = TopologyGraph()
        root = NodeSet('root')
        ns01 = NodeSet('nodes[0-1]')
        ns23 = NodeSet('nodes[2-3]')
        ns45 = NodeSet('nodes[4-5]')
        ns67 = NodeSet('nodes[6-7]')
        ns89 = NodeSet('nodes[8-9]')

        g.add_route(root, ns01)
        g.add_route(root, ns23 | ns45)
        self.assertRaises(TopologyError, g.add_route, ns23, ns23)
        self.assertRaises(TopologyError, g.add_route, ns45, root)
        g.add_route(ns23, ns67)
        g.add_route(ns67, ns89)
        self.assertRaises(TopologyError, g.add_route, ns89, ns67)
        self.assertRaises(TopologyError, g.add_route, ns89, ns89)
        self.assertRaises(TopologyError, g.add_route, ns89, ns23)

        ns_all = NodeSet('root,nodes[0-9]')
        for nodegroup in g.to_tree('root'):
            ns_all.difference_update(nodegroup.nodeset)
        self.assertEqual(len(ns_all), 0)

    def testInvalidRootNode(self):
        """test invalid root node specification"""
        g = TopologyGraph()
        ns0 = NodeSet('node[0-9]')
        ns1 = NodeSet('node[10-19]')
        g.add_route(ns0, ns1)
        self.assertRaises(TopologyError, g.to_tree, 'admin1')

    def testMultipleAdminGroups(self):
        """test topology with several admin groups"""
        ## -------------------
        # TODO : uncommenting following lines should not produce an error. This
        # is a valid topology!!
        # ----------
        tmpfile = tempfile.NamedTemporaryFile()
        tmpfile.write(b'[routes]\n')
        tmpfile.write(b'admin0: nodes[0-1]\n')
        #tmpfile.write(b'admin1: nodes[0-1]\n')
        tmpfile.write(b'admin2: nodes[2-3]\n')
        #tmpfile.write(b'admin3: nodes[2-3]\n')
        tmpfile.write(b'nodes[0-1]: nodes[10-19]\n')
        tmpfile.write(b'nodes[2-3]: nodes[20-29]\n')
        tmpfile.flush()
        parser = TopologyParser(tmpfile.name)

        ns_all = NodeSet('admin2,nodes[2-3,20-29]')
        ns_tree = NodeSet()
        tree = parser.tree('admin2')
        self.assertEqual(tree.inner_node_count(), 3)
        self.assertEqual(tree.leaf_node_count(), 10)
        for nodegroup in tree:
            ns_tree.add(nodegroup.nodeset)
        self.assertEqual(str(ns_all), str(ns_tree))

    def testTopologyGraphBigGroups(self):
        """test adding huge nodegroups in routes"""
        g = TopologyGraph()
        ns0 = NodeSet('nodes[0-10000]')
        ns1 = NodeSet('nodes[12000-23000]')
        g.add_route(ns0, ns1)
        self.assertEqual(g.dest(ns0), ns1)

        ns2 = NodeSet('nodes[30000-35000]')
        ns3 = NodeSet('nodes[35001-45000]')
        g.add_route(ns2, ns3)
        self.assertEqual(g.dest(ns2), ns3)

    def testNodeString(self):
        """test loading a linear string topology"""
        tmpfile = tempfile.NamedTemporaryFile()
        tmpfile.write(b'[routes]\n')

        # TODO : increase the size
        ns = NodeSet('node[0-10]')

        prev = 'admin'
        for n in ns:
            line = '%s: %s\n' % (prev, str(n))
            tmpfile.write(line.encode())
            prev = n
        tmpfile.flush()
        parser = TopologyParser(tmpfile.name)

        tree = parser.tree('admin')

        ns.add('admin')
        ns_tree = NodeSet()
        for nodegroup in tree:
            ns_tree.add(nodegroup.nodeset)
        self.assertEquals(ns, ns_tree)

    def testConfigurationParser(self):
        """test configuration parsing"""
        tmpfile = tempfile.NamedTemporaryFile()
        tmpfile.write(b'# this is a comment\n')
        tmpfile.write(b'[routes]\n')
        tmpfile.write(b'admin: nodes[0-1]\n')
        tmpfile.write(b'nodes[0-1]: nodes[2-5]\n')
        tmpfile.write(b'nodes[4-5]: nodes[6-9]\n')
        tmpfile.flush()
        parser = TopologyParser(tmpfile.name)

        parser.tree('admin')
        ns_all = NodeSet('admin,nodes[0-9]')
        ns_tree = NodeSet()
        for nodegroup in parser.tree('admin'):
            ns_tree.add(nodegroup.nodeset)
        self.assertEqual(str(ns_all), str(ns_tree))

    def testConfigurationParserCompatMain(self):
        """test configuration parsing (Main section compat)"""
        tmpfile = tempfile.NamedTemporaryFile()
        tmpfile.write(b'# this is a comment\n')
        tmpfile.write(b'[Main]\n')
        tmpfile.write(b'admin: nodes[0-1]\n')
        tmpfile.write(b'nodes[0-1]: nodes[2-5]\n')
        tmpfile.write(b'nodes[4-5]: nodes[6-9]\n')
        tmpfile.flush()
        parser = TopologyParser(tmpfile.name)

        parser.tree('admin')
        ns_all = NodeSet('admin,nodes[0-9]')
        ns_tree = NodeSet()
        for nodegroup in parser.tree('admin'):
            ns_tree.add(nodegroup.nodeset)
        self.assertEqual(str(ns_all), str(ns_tree))

    def testConfigurationShortSyntax(self):
        """test short topology specification syntax"""
        tmpfile = tempfile.NamedTemporaryFile()
        tmpfile.write(b'# this is a comment\n')
        tmpfile.write(b'[routes]\n')
        tmpfile.write(b'admin: nodes[0-9]\n')
        tmpfile.write(b'nodes[0-3,5]: nodes[10-19]\n')
        tmpfile.write(b'nodes[4,6-9]: nodes[30-39]\n')
        tmpfile.flush()
        parser = TopologyParser()
        parser.load(tmpfile.name)

        ns_all = NodeSet('admin,nodes[0-19,30-39]')
        ns_tree = NodeSet()
        for nodegroup in parser.tree('admin'):
            ns_tree.add(nodegroup.nodeset)
        self.assertEqual(str(ns_all), str(ns_tree))

    def testConfigurationLongSyntax(self):
        """test detailed topology description syntax"""
        tmpfile = tempfile.NamedTemporaryFile()
        tmpfile.write(b'# this is a comment\n')
        tmpfile.write(b'[routes]\n')
        tmpfile.write(b'admin: proxy\n')
        tmpfile.write(b'proxy: STA[0-1]\n')
        tmpfile.write(b'STA0: STB[0-1]\n')
        tmpfile.write(b'STB0: nodes[0-2]\n')
        tmpfile.write(b'STB1: nodes[3-5]\n')
        tmpfile.write(b'STA1: STB[2-3]\n')
        tmpfile.write(b'STB2: nodes[6-7]\n')
        tmpfile.write(b'STB3: nodes[8-10]\n')

        tmpfile.flush()
        parser = TopologyParser()
        parser.load(tmpfile.name)

        ns_all = NodeSet('admin,proxy,STA[0-1],STB[0-3],nodes[0-10]')
        ns_tree = NodeSet()
        tree = parser.tree('admin')
        self.assertEqual(tree.inner_node_count(), 8)
        self.assertEqual(tree.leaf_node_count(), 11)
        for nodegroup in tree:
            ns_tree.add(nodegroup.nodeset)
        self.assertEqual(str(ns_all), str(ns_tree))

    def testConfigurationParserDeepTree(self):
        """test a configuration that generates a deep tree"""
        tmpfile = tempfile.NamedTemporaryFile()
        tmpfile.write(b'# this is a comment\n')
        tmpfile.write(b'[routes]\n')
        tmpfile.write(b'admin: nodes[0-9]\n')

        levels = 15 # how deep do you want the tree to be?
        for i in range(0, levels*10, 10):
            line = 'nodes[%d-%d]: nodes[%d-%d]\n' % (i, i+9, i+10, i+19)
            tmpfile.write(line.encode())
        tmpfile.flush()
        parser = TopologyParser()
        parser.load(tmpfile.name)

        ns_all = NodeSet('admin,nodes[0-159]')
        ns_tree = NodeSet()
        tree = parser.tree('admin')
        self.assertEqual(tree.inner_node_count(), 151)
        self.assertEqual(tree.leaf_node_count(), 10)
        for nodegroup in tree:
            ns_tree.add(nodegroup.nodeset)
        self.assertEqual(str(ns_all), str(ns_tree))

    def testConfigurationParserBigTree(self):
        """test configuration parser against big propagation tree"""
        tmpfile = tempfile.NamedTemporaryFile()
        tmpfile.write(b'# this is a comment\n')
        tmpfile.write(b'[routes]\n')
        tmpfile.write(b'admin: ST[0-4]\n')
        tmpfile.write(b'ST[0-4]: STA[0-49]\n')
        tmpfile.write(b'STA[0-49]: nodes[0-10000]\n')
        tmpfile.flush()
        parser = TopologyParser()
        parser.load(tmpfile.name)

        ns_all = NodeSet('admin,ST[0-4],STA[0-49],nodes[0-10000]')
        ns_tree = NodeSet()
        tree = parser.tree('admin')
        self.assertEqual(tree.inner_node_count(), 56)
        self.assertEqual(tree.leaf_node_count(), 10001)
        for nodegroup in tree:
            ns_tree.add(nodegroup.nodeset)
        self.assertEqual(str(ns_all), str(ns_tree))

    def testConfigurationParserConvergentPaths(self):
        """convergent paths detection"""
        tmpfile = tempfile.NamedTemporaryFile()
        tmpfile.write(b'# this is a comment\n')
        tmpfile.write(b'[routes]\n')
        tmpfile.write(b'fortoy32: fortoy[33-34]\n')
        tmpfile.write(b'fortoy33: fortoy35\n')
        tmpfile.write(b'fortoy34: fortoy36\n')
        tmpfile.write(b'fortoy[35-36]: fortoy37\n')

        tmpfile.flush()
        parser = TopologyParser()
        self.assertRaises(TopologyError, parser.load, tmpfile.name)

    def testPrintingTree(self):
        """test printing tree"""
        tmpfile = tempfile.NamedTemporaryFile()
        tmpfile.write(b'[routes]\n')
        tmpfile.write(b'n0: n[1-2]\n')
        tmpfile.write(b'n1: n[10-49]\n')
        tmpfile.write(b'n2: n[50-89]\n')
        tmpfile.flush()
        parser = TopologyParser()
        parser.load(tmpfile.name)

        tree = parser.tree('n0')

        # In fact it looks like this:
        # ---------------------------
        # n0
        # |_ n1
        # |  |_ n[10-49]
        # |_ n2
        #    |_ n[50-89]
        # ---------------------------
        display_ref1 = 'n0\n|- n1\n|  `- n[10-49]\n`- n2\n   `- n[50-89]\n'
        display_ref2 = 'n0\n|- n2\n|  `- n[50-89]\n`- n1\n   `- n[10-49]\n'
        display = str(tree)
        self.assertTrue(display == display_ref1 or display == display_ref2)

        self.assertEquals(str(TopologyTree()), '<TopologyTree instance (empty)>')

    def testAddingInvalidChildren(self):
        """test detecting invalid children"""
        t0 = TopologyNodeGroup(NodeSet('node[0-9]'))
        self.assertRaises(AssertionError, t0.add_child, 'foobar')
        t1 = TopologyNodeGroup(NodeSet('node[10-19]'))

        t0.add_child(t1)
        self.assertEquals(t0.children_ns(), t1.nodeset)
        t0.add_child(t1)
        self.assertEquals(t0.children_ns(), t1.nodeset)

    def testRemovingChild(self):
        """test child removal operation"""
        t0 = TopologyNodeGroup(NodeSet('node[0-9]'))
        t1 = TopologyNodeGroup(NodeSet('node[10-19]'))

        t0.add_child(t1)
        self.assertEquals(t0.children_ns(), t1.nodeset)
        t0.clear_child(t1)
        self.assertEquals(t0.children_ns(), None)

        t0.clear_child(t1) # error discarded
        self.assertRaises(ValueError, t0.clear_child, t1, strict=True)

        t2 = TopologyNodeGroup(NodeSet('node[20-29]'))
        t0.add_child(t1)
        t0.add_child(t2)
        self.assertEquals(t0.children_ns(), t1.nodeset | t2.nodeset)
        t0.clear_children()
        self.assertEquals(t0.children_ns(), None)
        self.assertEquals(t0.children_len(), 0)

    def testStrConversions(self):
        """test str() casts"""
        t = TopologyNodeGroup(NodeSet('admin0'))
        self.assertEquals(str(t), '<TopologyNodeGroup (admin0)>')

        t = TopologyRoutingTable()
        r0 = TopologyRoute(NodeSet('src[0-9]'), NodeSet('dst[5-8]'))
        r1 = TopologyRoute(NodeSet('src[10-19]'), NodeSet('dst[15-18]'))

        self.assertEquals(str(r0), 'src[0-9] -> dst[5-8]')

        t.add_route(r0)
        t.add_route(r1)
        self.assertEquals(str(t), 'src[0-9] -> dst[5-8]\nsrc[10-19] -> dst[15-18]')

        g = TopologyGraph()
        # XXX: Actually if g is not empty other things will be printed out...
        self.assertEquals(str(g), '<TopologyGraph>\n')



# ClusterShell test suite
# Written by S. Thiell

"""Unit test for ClusterShell inter-Task msg"""

import threading
import time
import unittest

from ClusterShell.Task import *
from ClusterShell.Event import EventHandler


class TaskPortTest(unittest.TestCase):

    def tearDown(self):
        task_cleanup()

    def testPortMsg1(self):
        """test port msg from main thread to task"""

        TaskPortTest.got_msg = False

        # create task in new thread
        task = Task()

        class PortHandler(EventHandler):
            def ev_msg(self, port, msg):
                # receive msg
                assert msg == "toto"
                assert port.task.thread == threading.currentThread()
                TaskPortTest.got_msg = True
                port.task.abort()

        # create non-autoclosing port
        port = task.port(handler=PortHandler())
        task.resume()
        # send msg from main thread
        port.msg("toto")
        task_wait()
        self.assertTrue(TaskPortTest.got_msg)

    def testPortRemove(self):
        """test remove_port()"""

        class PortHandler(EventHandler):
            def ev_msg(self, port, msg):
                pass

        task = Task() # new thread
        port = task.port(handler=PortHandler(), autoclose=True)
        task.resume()
        task.remove_port(port)
        task_wait()

    def testPortClosed(self):
        """test port msg on closed port"""
        # test sending message to "stillborn" port
        self.port_msg_result = None

        # thread will wait a bit and send a port message
        def test_thread_start(port, test):
            time.sleep(0.5)
            test.port_msg_result = port.msg('foobar')

        class TestHandler(EventHandler):
            pass

        task = task_self()
        test_handler = TestHandler()
        task.timer(0.2, handler=test_handler, autoclose=False)
        port = task.port(handler=test_handler, autoclose=True)
        thread = threading.Thread(None, test_thread_start, args=(port, self))
        thread.setDaemon(True)
        thread.start()
        task.resume()
        task.abort(kill=True) # will remove_port()
        thread.join()
        self.assertEqual(self.port_msg_result, False) # test vs. None and True



"""
Unit test for StreamWorker
"""

import os
import unittest

from ClusterShell.Worker.Worker import StreamWorker, WorkerError
from ClusterShell.Task import task_self
from ClusterShell.Event import EventHandler


class StreamTest(unittest.TestCase):

    def run_worker(self, worker):
        """helper method to schedule and run a worker"""
        task_self().schedule(worker)
        task_self().run()

    def test_001_empty(self):
        """test empty StreamWorker"""
        # that makes no sense but well...
        # handler=None is supported by base Worker class
        self.run_worker(StreamWorker(handler=None))

    def test_002_pipe_readers(self):
        """test StreamWorker bound to several pipe readers"""

        streams = { "pipe1_reader": b"Some data to read from a pipe",
                    "stderr": b"Error data to read using special keyword stderr",
                    "pipe2_reader": b"Other data to read from another pipe",
                    "pipe3_reader": b"Cool data to read from a third pipe" }

        class TestH(EventHandler):
            def __init__(self, testcase):
                self.snames = set()
                self.testcase = testcase

            def ev_error(self, worker):
                # test that ev_error is called in case of 'stderr' stream name
                self.testcase.assertEqual(worker.current_sname, "stderr")
                self.recv_msg(worker.current_errmsg)

            def ev_read(self, worker):
                self.recv_msg(worker.current_msg)

            def recv_msg(self, msg):
                self.testcase.assertTrue(len(self.snames) < len(streams))
                self.testcase.assertEqual(streams[worker.current_sname], msg)
                self.snames.add(worker.current_sname)
                if len(self.snames) == len(streams):
                    # before finishing, try to add another pipe at
                    # runtime: this is NOT allowed
                    rfd, wfd = os.pipe()
                    self.testcase.assertRaises(WorkerError,
                        worker.set_reader, "pipe4_reader", rfd)
                    self.testcase.assertRaises(WorkerError,
                        worker.set_writer, "pipe4_writer", wfd)
                    os.close(rfd)
                    os.close(wfd)

        # create a StreamWorker instance bound to several pipes
        hdlr = TestH(self)
        worker = StreamWorker(handler=hdlr)

        for sname in streams.keys():
            rfd, wfd = os.pipe()
            worker.set_reader(sname, rfd)
            os.write(wfd, streams[sname])
            os.close(wfd)

        self.run_worker(worker)

        # check that all ev_read have been received
        self.assertEqual(set(("pipe1_reader", "pipe2_reader", "pipe3_reader",
                              "stderr")), hdlr.snames)

    def test_003_io_pipes(self):
        """test StreamWorker bound to pipe readers and writers"""

        # os.write -> pipe1 -> worker -> pipe2 -> os.read

        class TestH(EventHandler):
            def __init__(self, testcase):
                self.testcase = testcase
                self.worker = None
                self.pickup_count = 0
                self.hup_count = 0

            def ev_pickup(self, worker):
                self.pickup_count += 1

            def ev_read(self, worker):
                self.testcase.assertEqual(worker.current_sname, "pipe1")
                worker.write(worker.current_msg, "pipe2")

            def ev_timer(self, timer):
                # call set_write_eof on specific stream after some delay
                worker = self.worker
                self.worker = 'DONE'
                worker.set_write_eof("pipe2")

            def ev_hup(self, worker):
                # ev_hup called at the end (after set_write_eof is called)
                self.hup_count += 1
                self.testcase.assertEqual(self.worker, 'DONE')
                # no rc code should be set
                self.testcase.assertEqual(worker.current_rc, None)

        # create a StreamWorker instance bound to several pipes
        hdlr = TestH(self)
        worker = StreamWorker(handler=hdlr)
        hdlr.worker = worker

        rfd1, wfd1 = os.pipe()
        worker.set_reader("pipe1", rfd1)
        os.write(wfd1, b"Some data\n")
        os.close(wfd1)

        rfd2, wfd2 = os.pipe()
        worker.set_writer("pipe2", wfd2)

        timer1 = task_self().timer(1.0, handler=hdlr)
        self.run_worker(worker)
        self.assertEqual(os.read(rfd2, 1024), b"Some data")
        os.close(rfd2)
        # wfd2 should be closed by CS
        self.assertRaises(OSError, os.close, wfd2)
        # rfd1 should be closed by CS
        self.assertRaises(OSError, os.close, rfd1)
        # check pickup/hup
        self.assertEqual(hdlr.hup_count, 1)
        self.assertEqual(hdlr.pickup_count, 1)
        self.assertTrue(task_self().max_retcode() is None)

    def test_004_timeout_on_open_stream(self):
        """test StreamWorker with timeout set on open stream"""
        # Create worker set with timeout
        worker = StreamWorker(handler=None, timeout=0.5)

        # Create pipe stream
        rfd1, wfd1 = os.pipe()
        worker.set_reader("pipe1", rfd1, closefd=False)
        # Write some chars without line break (worst case)
        os.write(wfd1, b"Some data")
        # TEST: Do not close wfd1 to simulate open stream

        # Need to enable pipe1_msgtree
        task_self().set_default("pipe1_msgtree", True)
        self.run_worker(worker)

        # Timeout occured - read buffer should have been flushed
        self.assertEqual(worker.read(sname="pipe1"), b"Some data")

        # closefd was set, we should be able to close pipe fds
        os.close(rfd1)
        os.close(wfd1)

    def test_005_timeout_events(self):
        """test StreamWorker with timeout set (event based)"""
        class TestH(EventHandler):
            def __init__(self, testcase):
                self.testcase = testcase
                self.ev_pickup_called = False
                self.ev_read_called = False
                self.ev_hup_called = False
                self.ev_timeout_called = False

            def ev_pickup(self, worker):
                self.ev_pickup_called = True

            def ev_read(self, worker):
                self.ev_read_called = True
                self.testcase.assertEqual(worker.current_sname, "pipe1")
                self.testcase.assertEqual(worker.current_msg, b"Some data")

            def ev_hup(self, worker):
                # ev_hup is called but no rc code should be set
                self.ev_hup_called = True
                self.testcase.assertEqual(worker.current_rc, None)

            def ev_timeout(self, worker):
                self.ev_timeout_called = True

        hdlr = TestH(self)
        worker = StreamWorker(handler=hdlr, timeout=0.5)

        # Create pipe stream with closefd set (default)
        rfd1, wfd1 = os.pipe()
        worker.set_reader("pipe1", rfd1)
        # Write some chars without line break (worst case)
        os.write(wfd1, b"Some data")
        # TEST: Do not close wfd1 to simulate open stream

        self.run_worker(worker)
        self.assertTrue(hdlr.ev_timeout_called)
        self.assertTrue(hdlr.ev_read_called)
        self.assertTrue(hdlr.ev_pickup_called)
        self.assertTrue(hdlr.ev_hup_called)

        # rfd1 should be already closed by CS
        self.assertRaises(OSError, os.close, rfd1)
        os.close(wfd1)

    def test_006_worker_abort_on_written(self):
        """test StreamWorker abort on ev_written"""

        # This test creates a writable StreamWorker that will abort after the
        # first write, to check whether ev_written is generated in the right
        # place.

        class TestH(EventHandler):
            def __init__(self, testcase, rfd):
                self.testcase = testcase
                self.rfd = rfd
                self.check_written = 0

            def ev_written(self, worker, node, sname, size):
                self.check_written += 1
                self.testcase.assertEqual(os.read(self.rfd, 1024), b"initial")
                worker.abort()
                worker.abort()  # safe but no effect

        rfd, wfd = os.pipe()

        hdlr = TestH(self, rfd)
        worker = StreamWorker(handler=hdlr)

        worker.set_writer("test", wfd) # closefd=True
        worker.write(b"initial", "test")

        self.run_worker(worker)
        self.assertEqual(hdlr.check_written, 1)
        os.close(rfd)

    def test_007_worker_abort_on_written_eof(self):
        """test StreamWorker abort on ev_written (with EOF)"""

        # This test is similar to previous test test_006 but does
        # write() + set_write_eof().

        class TestH(EventHandler):
            def __init__(self, testcase, rfd):
                self.testcase = testcase
                self.rfd = rfd
                self.check_written = 0

            def ev_written(self, worker, node, sname, size):
                self.check_written += 1
                self.testcase.assertEqual(os.read(self.rfd, 1024), b"initial")
                worker.abort()
                worker.abort()  # safe but no effect

        rfd, wfd = os.pipe()

        hdlr = TestH(self, rfd)
        worker = StreamWorker(handler=hdlr)

        worker.set_writer("test", wfd) # closefd=True
        worker.write(b"initial", "test")
        worker.set_write_eof()

        self.run_worker(worker)
        self.assertEqual(hdlr.check_written, 1)
        os.close(rfd)

    def test_008_broken_pipe_on_write(self):
        """test StreamWorker with broken pipe on write()"""

        # This test creates a writable StreamWorker that will close the read
        # side of the pipe just after the first write to generate a broken
        # pipe error.

        class TestH(EventHandler):
            def __init__(self, testcase, rfd):
                self.testcase = testcase
                self.rfd = rfd
                self.check_hup = 0
                self.check_written = 0

            def ev_hup(self, worker):
                self.check_hup += 1

            def ev_written(self, worker, node, sname, size):
                self.check_written += 1
                self.testcase.assertEqual(os.read(self.rfd, 1024), b"initial")
                # close reader, that will stop the StreamWorker
                os.close(self.rfd)
                # The following write call used to raise broken pipe before
                # version 1.7.2.
                worker.write(b"final")

        rfd, wfd = os.pipe()

        hdlr = TestH(self, rfd)
        worker = StreamWorker(handler=hdlr)

        worker.set_writer("test", wfd) # closefd=True
        worker.write(b"initial", "test")

        self.run_worker(worker)
        self.assertEqual(hdlr.check_hup, 1)
        self.assertEqual(hdlr.check_written, 1)

    def test_009_worker_abort_on_close(self):
        """test StreamWorker abort() on closing worker"""

        class TestH(EventHandler):
            def __init__(self, testcase, rfd):
                self.testcase = testcase
                self.rfd = rfd
                self.check_close = 0

            def ev_close(self, worker, timedout):
                self.check_close += 1
                self.testcase.assertFalse(timedout)
                os.close(self.rfd)
                worker.abort()
                worker.abort()  # safe but no effect

        rfd, wfd = os.pipe()

        hdlr = TestH(self, rfd)
        worker = StreamWorker(handler=hdlr)

        worker.set_writer("test", wfd)  # closefd=True
        worker.write(b"initial", "test")
        worker.set_write_eof()

        self.run_worker(worker)
        self.assertEqual(hdlr.check_close, 1)



#
# Copyright (C) 2009-2015 CEA/DAM
#
# This file is part of ClusterShell.
#
# ClusterShell is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# ClusterShell is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with ClusterShell; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""
A ClusterShell Engine using epoll, an I/O event notification facility.

The epoll event distribution interface is available on Linux 2.6, and
has been included in Python 2.6.
"""

import errno
import select
import time

from ClusterShell.Engine.Engine import Engine, E_READ, E_WRITE
from ClusterShell.Engine.Engine import EngineNotSupportedError
from ClusterShell.Engine.Engine import EngineTimeoutException
from ClusterShell.Worker.EngineClient import EngineClientEOF


class EngineEPoll(Engine):
    """
    EPoll Engine

    ClusterShell Engine class using the select.epoll mechanism.
    """

    identifier = "epoll"

    def __init__(self, info):
        """
        Initialize Engine.
        """
        Engine.__init__(self, info)
        try:
            # get an epoll object
            self.epolling = select.epoll()
        except AttributeError:
            raise EngineNotSupportedError(EngineEPoll.identifier)

    def release(self):
        """Release engine-specific resources."""
        self.epolling.close()

    def _register_specific(self, fd, event):
        """Engine-specific fd registering. Called by Engine register."""
        if event & E_READ:
            eventmask = select.EPOLLIN
        else:
            assert event & E_WRITE
            eventmask = select.EPOLLOUT

        self.epolling.register(fd, eventmask)

    def _unregister_specific(self, fd, ev_is_set):
        """
        Engine-specific fd unregistering. Called by Engine unregister.
        """
        self._debug("UNREGSPEC fd=%d ev_is_set=%x"% (fd, ev_is_set))
        if ev_is_set:
            self.epolling.unregister(fd)

    def _modify_specific(self, fd, event, setvalue):
        """
        Engine-specific modifications after a interesting event change for
        a file descriptor. Called automatically by Engine set_events().
        For the epoll engine, it modifies the event mask associated to a file
        descriptor.
        """
        self._debug("MODSPEC fd=%d event=%x setvalue=%d" % (fd, event,
                                                            setvalue))
        if setvalue:
            self._register_specific(fd, event)
        else:
            self.epolling.unregister(fd)

    def runloop(self, timeout):
        """
        Run epoll main loop.
        """
        if not timeout:
            timeout = -1

        start_time = time.time()

        # run main event loop...
        while self.evlooprefcnt > 0:
            self._debug("LOOP evlooprefcnt=%d (reg_clifds=%s) (timers=%d)" % \
                    (self.evlooprefcnt, self.reg_clifds.keys(),
                     len(self.timerq)))
            try:
                timeo = self.timerq.nextfire_delay()
                if timeout > 0 and timeo >= timeout:
                    # task timeout may invalidate clients timeout
                    self.timerq.clear()
                    timeo = timeout
                elif timeo == -1:
                    timeo = timeout

                self._current_loopcnt += 1

                if timeo < 0:
                    poll_timeo = -1
                else:
                    poll_timeo = timeo
                evlist = self.epolling.poll(poll_timeo)

            except IOError as ex:
                # might get interrupted by a signal
                if ex.errno == errno.EINTR:
                    continue

            for fd, event in evlist:

                # get client instance
                client, stream = self._fd2client(fd)
                if client is None:
                    continue

                fdev = stream.evmask
                sname = stream.name

                # set as current processed stream
                self._current_stream = stream

                # check for poll error condition of some sort
                if event & select.EPOLLERR:
                    self._debug("EPOLLERR fd=%d sname=%s fdev=0x%x (%s)" % \
                                (fd, sname, fdev, client))
                    assert fdev & E_WRITE
                    self.remove_stream(client, stream)
                    self._current_stream = None
                    continue

                # check for data to read
                if event & select.EPOLLIN:
                    assert fdev & E_READ
                    assert stream.events & fdev, (stream.events, fdev)
                    self.modify(client, sname, 0, fdev)
                    try:
                        client._handle_read(sname)
                    except EngineClientEOF:
                        self._debug("EngineClientEOF %s %s" % (client, sname))
                        self.remove_stream(client, stream)
                        self._current_stream = None
                        continue

                # or check for end of stream (do not handle both at the same
                # time because handle_read() may perform a partial read)
                elif event & select.EPOLLHUP:
                    assert fdev & E_READ, "fdev 0x%x & E_READ" % fdev
                    self._debug("EPOLLHUP fd=%d sname=%s %s (%s)" % \
                                (fd, sname, client, client.streams))
                    self.remove_stream(client, stream)
                    self._current_stream = None
                    continue

                # check for writing
                if event & select.EPOLLOUT:
                    self._debug("EPOLLOUT fd=%d sname=%s %s (%s)" % \
                                (fd, sname, client, client.streams))
                    assert fdev & E_WRITE
                    assert stream.events & fdev, (stream.events, fdev)
                    self.modify(client, sname, 0, fdev)
                    client._handle_write(sname)

                self._current_stream = None

                # apply any changes occured during processing
                if client.registered:
                    self.set_events(client, stream)

            # check for task runloop timeout
            if timeout > 0 and time.time() >= start_time + timeout:
                raise EngineTimeoutException()

            # process clients timeout
            self.fire_timers()

        self._debug("LOOP EXIT evlooprefcnt=%d (reg_clifds=%s) (timers=%d)" % \
                (self.evlooprefcnt, self.reg_clifds, len(self.timerq)))

